# -*- coding: utf-8 -*-

__author__ = "Luis Barroso-Luque, Holden Parks"
__copyright__ = "Copyright 2016, CAMERA, LBL, ALS"
__credits__ = ["Ronald J Pandolfi", "Dinesh Kumar", "Singanallur Venkatakrishnan", "Luis Luque",
               "Holden Parks", "Alexander Hexemer"]
__license__ = ""
__version__ = "1.2.1"
__maintainer__ = "Ronald J Pandolfi"
__email__ = "ronpandolfi@lbl.gov"
__status__ = "Beta"


import inspect
import time
import os
import fabio
from collections import OrderedDict
from copy import deepcopy
from functools import partial
from modpkgs import yamlmod
import numpy as np
import pyqtgraph as pg
from PySide import QtCore, QtGui
from pyqtgraph.parametertree import Parameter, ParameterTree
import config
import reconpkg
import ui
from xicam.widgets import featurewidgets as fw
from .functionmanager import TestRangeDialog


class FunctionWidget(fw.FeatureWidget):
    """
    Subclass of FeatureWidget that defines attributes to show parameters to a given function and run the function
    with the given parameters. These should be used with the corresponding FunctionManager to run Tomography pipeline
    workflows


    Attributes
    ----------
    func_name : str
        Function name
    subfunc_name : str
        Specific function name
    input_functions : dict
        dictionary with keys being parameters of this function to be overriden, and values being a FunctionWidget
        whose function will override said parameter
    param_dict : dict
        Dictionary with parameter names and values
    _function : function
        Function object corresponding to the function represented by widget
    package : str
        name of package to which function belongs
    params : pyqtgraph.Parameter
        Parameter instance with function parameter exposed in UI
    missing_args : list of str
        Names of missing arguments not contained in param_dict

    Signals
    -------
    sigTestRange(QtGui.QWidget, str, tuple, dict)
        Emitted when parameter range test is requested. Emits the sending widget, a string with a message to log, and
        a tuple with the range values for the parameter


    Parameters
    ----------
    name : str
        generic name of function
    subname : str
        specific name of function under the generic name category
    package : python package
        package
    input_functions : dict, optional
        dictionary with keys being parameters of this function to be overriden, and values being a FunctionWidget
        whose function will override said parameter
    checkable : bool, optional
        bool to set the function to be toggled on and of when running constructed workflows
    closeable : bool, optional
        bool to set if the function can be deleted from the pipeline editor
    parent : QWidget
        parent of this FunctionWidget
    """

    sigTestRange = QtCore.Signal(QtGui.QWidget, str, tuple)

    # TODO perhaps its better to not pass in the package object but only a string, package object can be retrived from reconpkgs.packages dict
    def __init__(self, name, subname, package, input_functions=None, checkable=True, closeable=True, parent=None):
        self.name = name
        if name != subname:
            self.name += ' (' + subname + ')'
        super(FunctionWidget, self).__init__(self.name, checkable=checkable, closeable=closeable, parent=parent)

        self.func_name = name
        self.subfunc_name = subname
        self.input_functions = {}
        self.param_dict = {}
        self._function = getattr(package, config.names[self.subfunc_name][0])

        #perhaps unnecessary
        self.package = package.__name__

        # TODO have the children kwarg be passed to __init__
        self.params = Parameter.create(name=self.name, children=config.parameters[self.subfunc_name], type='group',
                                       readonly=False)

        self.form = ParameterTree(showHeader=False)
        self.form.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.form.customContextMenuRequested.connect(self.paramMenuRequested)
        self.form.setParameters(self.params, showTop=True)

        # Initialize parameter dictionary with keys and default values
        self.updateParamsDict()
        argspec = inspect.getargspec(self._function)
        default_argnum = len(argspec[3])
        self.param_dict.update({key : val for (key, val) in zip(argspec[0][-default_argnum:], argspec[3])})
        for key, val in self.param_dict.iteritems():
            if key in [p.name() for p in self.params.children()]:
                self.params.param(key).setValue(val)
                self.params.param(key).setDefault(val)

        # Create a list of argument names (this will most generally be the data passed to the function)
        self.missing_args = [i for i in argspec[0] if i not in self.param_dict.keys()]

        self.parammenu = QtGui.QMenu()
        action = QtGui.QAction('Test Parameter Range', self)
        action.triggered.connect(self.testParamTriggered)
        self.parammenu.addAction(action)

        self.previewButton.customContextMenuRequested.connect(self.menuRequested)
        self.menu = QtGui.QMenu()

        if input_functions is not None:
            for param, ipf in input_functions.iteritems():
                self.addInputFunction(param, ipf)

        # wire up param changed signals
        for param in self.params.children():
            param.sigValueChanged.connect(self.paramChanged)

        # change on/off icons
        icon = QtGui.QIcon()
        if checkable:
            icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_51.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
            icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_45.png"), QtGui.QIcon.Normal, QtGui.QIcon.On)
            self.previewButton.setCheckable(True)
        else:
            icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_45.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
            self.previewButton.setCheckable(False)
            self.previewButton.setChecked(True)

        self.previewButton.setIcon(icon)
        self.previewButton.setFlat(True)
        self.previewButton.setChecked(True)
        self.previewButton.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)

        self.allowed_types = {'str': str, 'int': int, 'float': float, 'bool': bool, 'unicode': unicode}

        # set widgets to never hide their subfunctions
        self.expand()

    def collapse(self):
        """
        This catches all "collapse" requests and passes them, so that FunctionWidgets are not collapsable
        """
        pass

    @property
    def enabled(self):
        """
        Boolean showing if the function widget is enabled (eye open/closed)
        """
        if self.previewButton.isChecked() or not self.previewButton.isCheckable():
            return True
        return False

    @enabled.setter
    def enabled(self, val):
        """
        Set enabled value by toggling the previewButton only if the widget is checkable
        """
        if val and self.previewButton.isCheckable():
            self.previewButton.setChecked(True)
        else:
            self.previewButton.setChecked(False)

    @property
    def exposed_param_dict(self):
        """
        Parameter dictionary with only the parameters that are shown in GUI
        """
        param_dict = {key: val for (key, val) in self.updated_param_dict.iteritems()
                      if key in [param.name() for param in self.params.children()]}
        return param_dict

    @property
    def partial(self):
        """
        Package up all parameters into a functools.partial
        """
        return partial(self._function, **self.updated_param_dict)

    @property
    def updated_param_dict(self):
        """
        Return the param dict of the FunctionWidget updated with proper types
        """

        if hasattr(self, 'defaults'):
            param_dict = {}
            for key, val in self.param_dict.iteritems():
                if type(val) is str and 'None' in val:
                    pass
                elif key in self.defaults.iterkeys():
                    arg_type = self.defaults[key]['type']
                    try:
                        param_dict[key] = arg_type(val)
                    except ValueError:
                        param_dict[key] = None
                else:
                    param_dict[key] = val
            return param_dict
        else:
            return self.param_dict


    @property
    def func_signature(self):
        """
        String for function signature. Hopefully this can eventually be used to save workflows as scripts :)
        """
        signature = str(self._function.__name__) + '('
        for arg in self.missing_args:
            signature += '{},'.format(arg)
        for param, value in self.param_dict.iteritems():
            signature += '{0}={1},'.format(param, value) if not isinstance(value, str) else \
                '{0}=\'{1}\','.format(param, value)
        return signature[:-1] + ')'


    def updateParamsDict(self):
        """
        Update the values of the parameter dictionary with the current values in UI
        """
        self.param_dict.update({param.name(): param.value() for param in self.params.children()})
        for p, ipf in self.input_functions.iteritems():
            ipf.updateParamsDict()

    def addInputFunction(self, parameter, functionwidget):
        """
        Add an input function widget

        Parameters
        ----------
        parameter : str
            Parameter name that will be overriden by return value of the input function
        functionwidget : FunctionWidget
            FunctionWidget representing the input function

        """

        if parameter in self.input_functions:  # Check to see if parameter already has input function
            if functionwidget.subfunc_name == self.input_functions[parameter].subfunc_name:
                raise AttributeError('Input function already exists')  # skip if the input function already exists
            self.removeInputFunction(parameter)  # Remove it if it will be replaced
        self.input_functions[parameter] = functionwidget
        self.addSubFeature(functionwidget)
        functionwidget.sigDelete.connect(lambda: self.removeInputFunction(parameter))

    def removeInputFunction(self, parameter):
        """
        Remove the input function for the given parameter

        Parameters
        ----------
        parameter : str
            Parameter name that will be overriden by return value of the input function

        """
        function = self.input_functions.pop(parameter)
        self.removeSubFeature(function)

    def paramChanged(self, param):
        """
        Slot connected to a pg.Parameter.sigChanged signal
        """
        if hasattr(self, 'defaults'):
            try:
                arg_type = self.defaults[param.name()]['type']
                try:
                    self.allowed_types[arg_type](param.value())
                    self.param_dict.update({param.name(): param.value()})
                except ValueError:
                    if param.value() == "None":
                        self.param_dict.update({param.name(): param.value()})
                    else:
                        param.setValue(self.param_dict[param.name()])
            except KeyError:
                self.param_dict.update({param.name(): param.value()})
        else:
            self.param_dict.update({param.name(): param.value()})

    def allReadOnly(self, boolean):
        """
        Make all parameter read only
        """
        for param in self.params.children():
            param.setReadonly(boolean)

    def menuRequested(self):
        """
        Context menu for functionWidget. Default is not menu.
        """
        pass

    def paramMenuRequested(self, pos):
        """
        Menus when a parameter in the form is right clicked
        """
        if self.form.currentItem().parent():
            self.parammenu.exec_(self.form.mapToGlobal(pos))

    def testParamTriggered(self):
        """
        Slot when a parameter range is clicked. Will emit the parameter name and the chosen range
        """
        param = self.form.currentItem().param
        if param.type() == 'int' or param.type() == 'float':
            start, end, step = None, None, None
            if 'limits' in param.opts:
                start, end = param.opts['limits']
                step = (end - start) / 3 + 1
            elif param.value() is not None:
                start, end, step = param.value() / 2, 4 * (param.value()) / 2, param.value() / 2
            test = TestRangeDialog(param.type(), (start, end, step))
        elif param.type() == 'list':
            test = TestListRangeDialog(param.opts['values'])
        else:
            return
        if test.exec_():
            self.sigTestRange.emit(self, param.name(), test.selectedRange())

class AnglesFunctionWidget(FunctionWidget):

    """
    Custom functionwidget to handle the calculation of projection angles from custom parameters (ang1 and range vs
    ang1 and ang2
    """

    @property
    def updated_param_dict(self):

        param_dict = {}
        param_dict['ang1'] = float(self.param_dict['ang1'] + self.param_dict['range'])
        param_dict['ang2'] = float(self.param_dict['ang1'])
        param_dict['nang'] = self.param_dict['nang']

        return param_dict

class NormalizeFunctionWidget(FunctionWidget):

    """
    Custom functionwidget class for normalization functions, particularly for the special case 'tomopy.normalize_roi'

    Attributes
    ----------
    selection_roi: pyqtgraph.ROI
        ROI to select normalization area

    Signals
    -------
    sigROIAdded(pyqtgraph.ROI)
        emitted when user adds an ROI to projection stack
    sigROIChanged
        emitted when ROI changes dimensions or location
    """

    sigROIAdded = QtCore.Signal(pg.ROI)
    sigROIChanged = QtCore.Signal()

    def __init__(self, name, subname, package, input_functions=None, checkable=True, closeable=True, parent=None):
        super(NormalizeFunctionWidget, self).__init__(name=name, subname=subname, package=package,
                                        input_functions=input_functions, checkable=checkable, closeable=closeable,
                                        parent=parent)

        if subname == 'ROI':
            self.selection_roi = pg.ROI([0, 0], [200, 200])
            self.params.child('Select roi').sigActivated.connect(self.setRoi)
            # self.previewButton.focusInEvent = lambda x: self.focusInEvent(x)
            # self.previewButton.focusOutEvent = lambda x: self.focusOutEvent(x)

    @property
    def updated_param_dict(self):
        """
        Returns usual param dict unless the function is normalizing from ROI
        """
        param_dict = FunctionWidget.updated_param_dict.fget(self)
        if self.subfunc_name == 'ROI':
            param_dict.update({'roi': [self.params.child('p11').value(), self.params.child('p12').value(),
                                  self.params.child('p21').value(), self.params.child('p22').value()]})
            for param in ['p11', 'p12', 'p21', 'p22', 'Select roi']:
                try:
                    param_dict.pop(param)
                except KeyError:
                    pass
        return param_dict

    # TODO : make this code work. It should hide the ROI when you click away from the functionwidget and show it
    # TODO : again when you click back onto it

    # def focusInEvent(self, event):
    #     if hasattr(self, 'selection_roi'):
    #         self.selection_roi.show()
    #
    # def focusOutEvent(self, event):
    #     if hasattr(self, 'selection_roi'):
    #         self.selection_roi.hide()


    def setRoi(self):
        """
        Adds pyqtgraph.ROI to projection images so user can select ROI from which to normalize
        """

        if self.selection_roi:
            del(self.selection_roi)

        self.selection_roi = pg.ROI([0, 0], [200, 200], pen=QtGui.QPen('y'), removable=True)
        self.selection_roi.addScaleHandle([1, 1], [0, 0])
        self.selection_roi.addScaleHandle([0, 0], [1, 1])
        self.sigROIAdded.emit(self.selection_roi)

        self.params.child('p11').setValue(0)
        self.params.child('p12').setValue(200)
        self.params.child('p21').setValue(0)
        self.params.child('p22').setValue(200)

    def adjustParams(self, func_manager):
        """
        Adjusts parameters in functionwidget based on location of ROI. Also ensures values do not become negative
        or larger than image size
        """

        x_max, y_max = None, None
        for feature in func_manager.features:
            if 'Reader' in feature.name:
                x_max = feature.params.child('end_width').defaultValue()
                y_max = feature.params.child('end_sinogram').defaultValue()

        pos = self.selection_roi.pos()
        size = self.selection_roi.size()

        # there has to be a better way to do this
        if 0 < pos[0] < x_max:
            x1 = int(pos[0])
        elif pos[0] < 0:
            x1 = 0
        else:
            x1 = x_max

        if 0 < pos[0] + size[0] < x_max:
            x2 = int(pos[0] + size[0])
        elif pos[0] + size[0] < 0:
            x2 = 0
        else:
            x2 = x_max

        if 0 < pos[1] < y_max:
            y1 = int(pos[1])
        elif pos[1] < 0:
            y1 = 0
        else:
            y1 = y_max

        if 0 < pos[1] + size[1] < y_max:
            y2 = int(pos[1] + size[1])
        elif pos[1] + size[1] < 0:
            y2 = 0
        else:
            y2 = y_max

        self.params.child('p11').setValue(x1)
        self.params.child('p12').setValue(x2)
        self.params.child('p21').setValue(y1)
        self.params.child('p22').setValue(y2)



class TomoPyReconFunctionWidget(FunctionWidget):
    """
    Subclass of FunctionWidget used for Tomopy recon functions. Allows adding input functions for center of rotation
    detection. And has a default Projection Angles function to provide the theta parameter to the function.

    Parameters
    ----------
    name : str
        generic name of function
    subname : str
        specific name of function under the generic name category
    package : python package
        package
    """

    def __init__(self, name, subname, package):

        self.packagename = package.__name__
        self.input_functions = {'theta': AnglesFunctionWidget('Projection Angles', 'Projection Angles', closeable=False,
                                                  package=reconpkg.packages['tomopy'], checkable=False),
                                'center': FunctionWidget('Center Detection', 'Phase Correlation', closeable=True,
                                                  package=reconpkg.packages['tomopy'])}
        super(TomoPyReconFunctionWidget, self).__init__(name, subname, package, input_functions=self.input_functions,
                                                        checkable=False)
        # Fill in the appropriate 'algorithm' keyword
        self.param_dict['algorithm'] = subname.lower()
        self.submenu = QtGui.QMenu('Input Function')
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_39.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.submenu.setIcon(icon)
        ui.build_function_menu(self.submenu, config.funcs['Input Functions'][name], config.names,
                               self.addCenterDetectFunction)
        self.menu.addMenu(self.submenu)

        if subname in config.function_defaults.keys():
            self.defaults = config.function_defaults[subname]
            for key in self.defaults.iterkeys():
                val = self.defaults[key]['default']
                self.param_dict.update({key: val})
                if key in [p.name() for p in self.params.children()]:
                    self.params.child(key).setValue(val)
                    self.params.child(key).setDefault(val)

    @property
    def partial(self):
        """
        Package up all parameters into a functools.partial
        """
        return partial(self._function, **self.updated_param_dict)

    @property
    def updated_param_dict(self):
        """
        Return the param dict of the FunctionWidget updated with proper types
        """
        param_dict = {}
        for key, val in self.param_dict.iteritems():
            if key in self.defaults.iterkeys():
                arg_type = self.defaults[key]['type']
                try:
                    param_dict[key] = self.allowed_types[arg_type](val)
                except ValueError:
                    param_dict[key] = None
            else:
                param_dict[key] = val
        # 'cutoff' and 'order' are not passed into the recon function in the correct format
        if 'cutoff' in param_dict.keys() and 'order' in param_dict.keys():
            param_dict['filter_par'] = list((param_dict.pop('cutoff'), param_dict.pop('order')))
        return param_dict


    def addCenterDetectFunction(self, name, subname, package=reconpkg.packages['tomopy']):
        """
        Used to add a center detection input function from the FunctionWidgets context menu

        Parameters
        ----------
        name : str
            generic name of function
        subname : str
            specific name of function under the generic name category
        package : python package
            package where function is defined
        """
        try:
            self.addInputFunction('center', FunctionWidget(name, subname, package=package))
        except AttributeError:
            pass

    def setCenterParam(self, value):
        """
        Sets the center parameter for the FunctionWidget to give value
        """
        self.params.child('center').setValue(value)
        self.params.child('center').setDefault(value)

    def menuRequested(self, pos):
        """
        Slot when menu is requested from FeatureWidget ui (ie to add input functions)
        """
        self.menu.exec_(self.previewButton.mapToGlobal(pos))


class TomoCamReconFuncWidget(TomoPyReconFunctionWidget):
    """
    Subclass of tomopy FunctionWidget used for TomoCam recon functions. Necessary to load keywords into list, as
    required by TomoCam API
    """

    @property
    def updated_param_dict(self):
        param_dict = dict(self.param_dict)
        input_params = {}
        input_params['gpu_device'] = 0
        input_params['oversamp_factor'] = param_dict['oversamp_factor']
        param_dict.pop('oversamp_factor')

        if 'gridrec'in param_dict['algorithm']:
            input_params['fbp_filter_param'] = param_dict['cutoff']
            param_dict.pop('cutoff')
        else:
            input_params['num_iter'] = param_dict['num_iter']
            param_dict.pop('num_iter')

        param_dict['input_params'] = input_params
        return param_dict




class AstraReconFuncWidget(TomoPyReconFunctionWidget):
    """
    Subclass of FunctionWidget used for Astra recon functions using Tomopy's astra wrapper
    """
    def __init__(self, name, subname, package):
        super(AstraReconFuncWidget, self).__init__(name, subname, reconpkg.packages['tomopy'])
        # self.tomopy_args = {'center':'center', 'sinogram_order': False, 'nchunk': None,
        #                     'init_recon': None, 'tomo': 'tomo', 'theta': 'theta'}
        self.tomopy_args = ['center', 'sinogram_order', 'nchunk', 'init_recon', 'tomo', 'theta',
                            'ncore', 'algorithm']


        self.param_dict['algorithm'] = reconpkg.packages['tomopy'].astra
        self.param_dict['options'] = {}
        self.param_dict['options']['method'] = subname.replace(' ', '_')
        if 'CUDA' in subname:
            self.param_dict['options']['proj_type'] = 'cuda'
        else:
            self.param_dict['options']['proj_type'] = 'linear'

        if subname in config.function_defaults.keys():
            self.defaults = config.function_defaults[subname]
            for key in self.defaults.iterkeys():
                val = self.defaults[key]['default']
                self.param_dict.update({key: val})
                if key in [p.name() for p in self.params.children()]:
                    self.params.child(key).setValue(val)
                    self.params.child(key).setDefault(val)



    @property
    def updated_param_dict(self):
        """
        Return the param dict of the FunctionWidget reorganized for proper execution as tomopy function
        """

        # copy optional params to the options/extra_options subdictionaries
        param_dict = {}
        param_dict['options'] = self.param_dict['options']
        param_dict['options']['extra_options']={}
        for key, val in self.param_dict.iteritems():
            if key in self.tomopy_args:
                if type(val) == str and 'None' in val:
                    param_dict[key] = None
                elif key in self.defaults.iterkeys():
                    arg_type = self.defaults[key]['type']
                    param_dict[key] = self.allowed_types[arg_type](val)
                else:
                    param_dict[key] = val
            elif key == 'options' or (type(val) is str and 'None' in val):
                pass
            elif 'num_iter' in key:
                param_dict['options']['num_iter'] = self.param_dict['num_iter']
            elif key in config.function_defaults['Astra'][self.subfunc_name].iterkeys():
                param_dict['options']['extra_options'][key] = val
            else:
                pass

        # get rid of extra_options if there are none
        if len(param_dict['options']['extra_options']) < 1:
            param_dict['options'].pop('extra_options')

        # if 'astra' in algorithm name, remove it
        if 'astra' in param_dict['options']['method']:
            param_dict['options']['method'] = param_dict['options']['method'].split('_astra')[0]

        return param_dict

class MaskFunctionWidget(FunctionWidget):
    """
    Subclass of FunctionWidget for functions with masks.
    """

    def __init__(self, name, subname, package, checkable=True, closeable=True, parent=None):
        super(MaskFunctionWidget, self).__init__(name, subname, package, checkable=checkable, closeable=closeable,
                                                 parent=parent)
        self.params.child('Browse mask image').sigActivated.connect(self.setBrowse)
        self.params.child('mask').sigValueChanged.connect(self.showOptions)
        self.params.child('mask path').hide()

    def showOptions(self):
        """
        Slot to receive signal when mask is added to function parameters. Shows mask path when this parameter is
        filled, otherwise hides that parameter
        """

        if self.params.child('mask').value() == 'custom mask':
            self.params.child('mask path').show()
        else:
            self.params.child('mask path').hide()
            if self.params.child('mask').value() == 'StructuredElementL':
                self.params.child('L').show()
            else:
                self.params.child('L').hide()

    def setBrowse(self):
        """
        Opens a QtGui.QFileDialog to get path to mask
        """
        path = str(QtGui.QFileDialog.getOpenFileName(None, 'Choose mask image')[0])
        try:
            self.mask_image = fabio.open(path).data
            self.params.param('mask path').show()
            self.params.param('mask path').setValue(path)
            self.params.param('mask').setValue('custom mask')
        except IOError:
            pass

    @property
    def updated_param_dict(self):
        param_dict = {}
        if self.params.param('mask').value() == 'custom mask':
            param_dict['mask'] = self.mask_image
        else:
            param_dict['mask'] = self.params.param('mask').value()
        param_dict['image'] = self.param_dict['image']
        param_dict['L'] = self.param_dict['L']
        param_dict['platform'] = self.param_dict['platform']
        return param_dict

class ReadFunctionWidget(FunctionWidget):
    """
    Subclass of FunctionWidget for reader functions.
    """
    def __init__(self, name, subname, package):
        super(ReadFunctionWidget, self).__init__(name, subname, package, checkable=False,)

    @property
    def sinograms(self):
        return (self.params.child('start_sinogram').value(), self.params.child('end_sinogram').value(),
                self.params.child('step_sinogram').value())

    @property
    def projections(self):
        return (self.params.child('start_projection').value(), self.params.child('end_projection').value(),
                self.params.child('step_projection').value())

    @property
    def sino_chunk(self):
        return self.params.child('sinograms_per_chunk').value()

    @property
    def proj_chunk(self):
        return self.params.child('projections_per_chunk').value()

    @property
    def width(self):
        return (self.params.child('start_width').value(), self.params.child('end_width').value(),
                self.params.child('step_width').value())


class WriteFunctionWidget(FunctionWidget):
    """
    Subclass of FunctionWidget for write functions to have a Browse button to use a QFileDialog for setting 'fname'
    parameter

    Parameters
    ----------
    name : str
        generic name of function
    subname : str
        specific name of function under the generic name category
    package : python package
        package where function is defined

    Attributes
    ----------
    parent : str
        string of the parent folder holding data set
    folder : str
        string of new folder to be written into
    file : str
        file names for reconstruction to write to
    """

    sigPipelineChanged = QtCore.Signal()

    def __init__(self, name, subname, package):
        super(WriteFunctionWidget, self).__init__(name, subname, package)
        # self.params.child('Browse').sigActivated.connect(
        #     lambda: self.params.child('parent folder').setValue( str(QtGui.QFileDialog.getSaveFileName(None,
        #     'Save reconstruction as', self.params.child('parent folder').value())[0])))

        self.parent = self.params.param('parent folder')
        self.folder = self.params.param('folder name')
        self.file = self.params.param('file name')
        self.params.child('Browse').sigActivated.connect(self.setBrowse)

        # connect signals to change full file name whenever smaller names are changed
        self.parent.sigValueChanged.connect(self.pathChanged)
        self.folder.sigValueChanged.connect(self.pathChanged)
        self.file.sigValueChanged.connect(self.pathChanged)

        # set full file name as read only
        # this line causes the 'invalid keyword argument - readonly' error during a reconstructin preview
        self.params.param('fname').setReadonly()

        # self.fname.sigValueChanged.connect(self.fileChanged)
        # self.fname.sigValueChanged.connect(self.folderChanged)
        # self.fname.sigValueChanged.connect(self.parentChanged)

    def setBrowse(self):
        """
        Uses result of browse button in 'parent folder' and 'folder name' fields to get write-out name
        """

        path = str(QtGui.QFileDialog.getSaveFileName(None, 'Save reconstruction as', dir=self.parent.value())[0])
        try:
            folder = path.split('/')[-1]
            parent = path.split(folder)[0]

            self.parent.setValue(parent)
            self.folder.setValue(folder)
        except ValueError:
            pass


    def pathChanged(self):
        """
        Slot to receive signal when one of parent/folder attributes is changed. This function adjusts the final
        'fname' parameter accordinly
        """

        self.params.param('fname').setValue(os.path.join(self.parent.value(), self.folder.value(),
                                                         self.file.value()))




    def updateParamsDict(self):
        """
        Overrides parent method of updateParamsDict to avoid adding the 'Browse' action as a parameter
        """


        self.param_dict.update({param.name(): param.value() for param in self.params.children()
                                if param.name() != 'Browse'})  # skip the Browse parameter!
        for p, ipf in self.input_functions.iteritems():
            ipf.updateParamsDict()

class CORSelectionWidget(QtGui.QWidget):

    cor_detection_funcs = ['Phase Correlation', 'Vo', 'Nelder-Mead']
    sigCORFuncChanged = QtCore.Signal(str, QtGui.QWidget)

    def __init__(self, subname='Phase Correlation', parent=None):
        super(CORSelectionWidget, self).__init__(parent=parent)

        self.layout = QtGui.QVBoxLayout()
        self.function = FunctionWidget(name="Center Detection", subname=subname,
                                package=reconpkg.packages[config.names[subname][1]])
        self.params = pg.parametertree.Parameter.create(name=self.function.name,
                                             children=config.parameters[self.function.subfunc_name], type='group')

        self.param_tree = pg.parametertree.ParameterTree()
        self.param_tree.setMinimumHeight(200)
        self.param_tree.setMinimumWidth(200)
        self.param_tree.setParameters(self.params, showTop=False)
        for key, val in self.function.param_dict.iteritems():
            if key in [p.name() for p in self.params.children()]:
                self.params.child(key).setValue(val)
                self.params.child(key).setDefault(val)

        self.method_box = QtGui.QComboBox()
        self.method_box.currentIndexChanged.connect(self.changeFunction)
        for item in self.cor_detection_funcs:
            self.method_box.addItem(item)
        self.method_box.currentIndexChanged.connect(self.corFuncChanged)

        label = QtGui.QLabel('COR detection function: ')
        method_layout = QtGui.QHBoxLayout()
        method_layout.addWidget(label)
        method_layout.addWidget(self.method_box)

        self.layout.addLayout(method_layout)
        self.layout.addWidget(self.param_tree)
        self.setLayout(self.layout)

    def corFuncChanged(self, index):
        """
        Slot to receive signal when self.method_box changes values. Emits new COR function name to be connected to
        a slot in FunctionManager
        """

        self.sigCORFuncChanged.emit(self.cor_detection_funcs[index], self)

    def changeFunction(self, index):
        """
        Slot to receive signal when self.method_box changes values. Changes COR function.
        """
        subname = self.method_box.itemText(index)
        self.layout.removeWidget(self.param_tree)

        self.function = FunctionWidget(name="Center Detection", subname=subname,
                                package=reconpkg.packages[config.names[subname][1]])
        self.params = pg.parametertree.Parameter.create(name=self.function.name,
                                             children=config.parameters[self.function.subfunc_name], type='group')
        self.param_tree = pg.parametertree.ParameterTree()
        self.param_tree.setMinimumHeight(200)
        self.param_tree.setMinimumWidth(200)
        self.param_tree.setParameters(self.params,showTop = False)
        for key, val in self.function.param_dict.iteritems():
            if key in [p.name() for p in self.params.children()]:
                self.params.child(key).setValue(val)
                self.params.child(key).setDefault(val)

        self.layout.addWidget(self.param_tree)
        self.setLayout(self.layout)


import base
from PySide import QtGui
import os
from xicam import xglobals

import widgets


class MOTDPlugin(base.plugin):
    name = 'MOTD'
    hidden = True


    def __init__(self, *args, **kwargs):
        self.rightwidget = None

        try:
            from PySide import QtWebKit
        except ImportError:
            pass
        else:
            self.centerwidget = webview = QtWebKit.QWebView()
            webview.load('MOTD.html')

        super(MOTDPlugin, self).__init__(*args, **kwargs)

    # Defer methods to other plugins

    def openSelected(self, *args, **kwargs):
        xglobals.plugins['Timeline'].instance.activate()
        xglobals.plugins['Timeline'].instance.openSelected(*args, **kwargs)

    def openfiles(self, *args, **kwargs):
        xglobals.plugins['Viewer'].instance.activate()
        xglobals.plugins['Viewer'].instance.openfiles(*args, **kwargs)

    def calibrate(self):
        xglobals.plugins['Viewer'].instance.activate()
        xglobals.plugins['Viewer'].instance.calibrate()

    def currentImage(self):
        xglobals.plugins['Viewer'].instance.activate()
        xglobals.plugins['Viewer'].instance.currentImage()


from PySide import QtGui
import sys
import base
import viewer


class LibraryPlugin(base.plugin):
    name = 'Library'

    def __init__(self, *args, **kwargs):


        if sys.platform == 'win32':
            libraryview = LibraryWidget(self, 'C://')
        else:
            libraryview = LibraryWidget(self, pipeline.pathtools.getRoot())

        self.centerwidget = libraryview

        self.rightwidget = None

        self.featureform = None

        self.bottomwidget = None

        self.toolbar = None

        super(plugin, self).__init__(*args, **kwargs)

        # self.centerwidget.sigOpenFile.connect(viewer.plugininstance.openfiles)


from PySide import QtGui
from PySide import QtCore
from PySide.QtCore import Qt
import os
import pipeline
import numpy as np


class LibraryWidget(QtGui.QScrollArea):
    sigOpenFile = QtCore.Signal(str)

    def __init__(self, *args, **kwargs):
        super(LibraryWidget, self).__init__()
        self.setWidgetResizable(True)
        self.setFocusPolicy(Qt.NoFocus)
        l = librarylayout(*args, **kwargs)
        w = QtGui.QWidget()
        w.setLayout(l)
        w.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding)
        self.setWidget(w)
        l.sigOpenFile.connect(self.sigOpenFile)


class FlowLayout(QtGui.QLayout):
    """
    A grid layout that wraps and adjusts to its items
    """

    def __init__(self, parent=None, margin=5, spacing=-1):
        super(FlowLayout, self).__init__(parent)

        self.margin = margin
        self.setSpacing(spacing)

        self.itemList = []


    def __del__(self):
        item = self.takeAt(0)
        while item:
            item = self.takeAt(0)

    def addItem(self, item):
        self.itemList.append(item)

    def count(self):
        return len(self.itemList)

    def itemAt(self, index):
        if index >= 0 and index < len(self.itemList):
            return self.itemList[index]

        return None

    def takeAt(self, index):
        if index >= 0 and index < len(self.itemList):
            return self.itemList.pop(index)

        return None

    def expandingDirections(self):
        return Qt.Orientations(Qt.Orientation(0))

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width):
        height = self.doLayout(QtCore.QRect(0, 0, width, 0), True)
        return height

    def setGeometry(self, rect):
        super(FlowLayout, self).setGeometry(rect)
        self.doLayout(rect, False)

    def sizeHint(self):
        return self.minimumSize()

    def minimumSize(self):
        size = QtCore.QSize()

        for item in self.itemList:
            size = size.expandedTo(item.minimumSize())

        size += QtCore.QSize(2 * self.margin, 2 * self.margin)
        return size

    def doLayout(self, rect, testOnly):
        x = rect.x()
        y = rect.y()
        lineHeight = 0

        for item in self.itemList:
            wid = item.widget()
            spaceX = self.spacing() + wid.style().layoutSpacing(QtGui.QSizePolicy.PushButton,
                                                                QtGui.QSizePolicy.PushButton,
                                                                Qt.Horizontal)
            spaceY = self.spacing() + wid.style().layoutSpacing(QtGui.QSizePolicy.PushButton,
                                                                QtGui.QSizePolicy.PushButton,
                                                                Qt.Vertical)
            nextX = x + item.sizeHint().width() + spaceX
            if nextX - spaceX > rect.right() and lineHeight > 0:
                x = rect.x()
                y = y + lineHeight + spaceY
                nextX = x + item.sizeHint().width() + spaceX
                lineHeight = 0

            if not testOnly:
                item.setGeometry(QtCore.QRect(QtCore.QPoint(x, y), item.sizeHint()))

            x = nextX
            lineHeight = max(lineHeight, item.sizeHint().height())

        return y + lineHeight - rect.y()


class librarylayout(FlowLayout):
    """
    Extend the flow layout to fill it with thumbwidgetitems representing files/folders
    """

    sigOpenFile = QtCore.Signal(str)

    def __init__(self, parentwindow, path='/Volumes'):
        super(librarylayout, self).__init__()
        self.parent = None
        self.parentwindow = parentwindow
        self.populate(path)

    def chdir(self, path):
        self.clear()
        self.populate(os.path.normpath(path))

    def populate(self, path):
        self.parent = QtCore.QDir()
        self.parent.cd(path)

        dir = QtCore.QDir(self.parent)
        entries = dir.entryList()

        for entry in entries:
            # print fileinfo.fileName()
            if not (entry == '..' and os.path.normpath(path) == pipeline.pathtools.getRoot()) and not entry == '.':
                nameoverride = entry if entry == '..' else None
                w = thumbwidgetitem(os.path.join(path, entry), parentwindow=self.parentwindow,
                                    nameoverride=nameoverride)
                self.addWidget(w)
                w.sigOpenFile.connect(self.sigOpenFile)
                w.sigChangeRoot.connect(self.chdir)


    def clear(self):
        while self.count():
            item = self.takeAt(0)
            widget = item.widget()
            if widget is not None:
                widget.deleteLater()


class thumbwidgetitem(QtGui.QFrame):
    """
    A single icon representing a file/folder that can be accessed/opened
    """
    sigChangeRoot = QtCore.Signal(str)
    sigOpenFile = QtCore.Signal(str)


    def __init__(self, path, parentwindow, nameoverride=None):
        path = os.path.normpath(path)

        self.foldericon = QtGui.QImage()
        self.foldericon.load('xicam/gui/GenericFolderIcon.png')

        self.fileicon = QtGui.QImage()
        self.fileicon.load('xicam/gui/post-360412-0-09676400-1365986245.png')

        print 'Library widget generated for ' + path
        super(thumbwidgetitem, self).__init__()
        self.parentwindow = parentwindow
        self.setObjectName('thumb')
        desiredsize = QtCore.QSize(160, 200)

        self.setFixedSize(desiredsize)
        self.setAutoFillBackground(True)
        self.setFocusPolicy(Qt.StrongFocus)

        self.layout = QtGui.QVBoxLayout(self)  # .frame

        self.path = path
        # print path
        dimg = pipeline.loader.diffimage(filepath=self.path)
        self.image = QtGui.QImage()
        # print os.path.splitext(path)[1]
        if os.path.isdir(path):
            self.image = self.foldericon
        elif os.path.splitext(path)[1] in pipeline.loader.acceptableexts:


            try:
                self.thumb = np.rot90(np.log(dimg.thumbnail * (dimg.thumbnail > 0) + (dimg.thumbnail < 1))).copy()
                print 'thumbmax:', np.max(self.thumb)
                self.thumb *= 255. / np.max(self.thumb)



                # if self.thumb is not None:
                self.image = QtGui.QImage(self.thumb.astype(np.uint8), self.thumb.shape[1], self.thumb.shape[0],
                                          self.thumb.shape[1],
                                          QtGui.QImage.Format_Indexed8)
            except Exception as ex:
                print ex.message

                self.image = self.fileicon


        else:
            self.image = self.fileicon

        image_label = ScaledLabel(self.image)
        image_label.setAlignment(Qt.AlignHCenter)
        self.layout.addWidget(image_label)

        self.namelabel = QtGui.QLabel(os.path.basename(path) if nameoverride is None else nameoverride)
        self.namelabel.setAlignment(Qt.AlignHCenter)
        self.namelabel.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Maximum)

        line = QtGui.QFrame()
        line.setGeometry(QtCore.QRect(0, 0, desiredsize.width() * 9 / 10, 3))
        line.setFrameShape(QtGui.QFrame.HLine)
        line.setFrameShadow(QtGui.QFrame.Sunken)
        self.layout.addWidget(line)

        self.layout.addWidget(self.namelabel)
        self.setLayout(self.layout)

    def enterEvent(self, *args, **kwargs):
        # self.frame.setFrameStyle(QFrame.Raised)
        pass

    # def leaveEvent(self, *args, **kwargs):
    # self.frame.setFrameStyle(QFrame.Plain)

    # def mousePressEvent(self, *args, **kwargs):
    # self.frame.setFrameStyle(QFrame.Sunken)

    def mouseDoubleClickEvent(self, *args, **kwargs):
        if os.path.isdir(self.path):
            self.sigChangeRoot.emit(self.path)
        else:
            self.sigOpenFile.emit(self.path)


class ScaledLabel(QtGui.QLabel):
    """
    A label that scales with the constrained dimension; for thumbnails
    """

    def __init__(self, image):
        super(ScaledLabel, self).__init__()
        self._pixmap = QtGui.QPixmap.fromImage(image)
        self._pixmap = QtGui.QPixmap(self._pixmap)

    def resizeEvent(self, event):
        self.setPixmap(self._pixmap.scaled(
            self.width(), self.height(),
            Qt.KeepAspectRatio))


#import nexpy.api.nexus as nx

import numpy as nx

import numpy as np
import scipy.ndimage
from PySide import QtCore
from xicam import debugtools, dialogs
import multiprocessing
import time
import os
from PIL import Image
from fabio import edfimage, tifimage
import scipy.misc
import msg


class nexusmerger(QtCore.QThread):
    def __init__(self, *args, **kwargs):
        super(nexusmerger, self).__init__()
        self.args = args
        self.kwargs = kwargs

    def run(self):
        p = multiprocessing.Process(target=mergenexus, kwargs=self.kwargs)
        self.job = p
        p.start()


def mergenexus(**kwargs):
    path = kwargs['path']

    newfile = not os.path.isfile(path)
    if newfile:
        nxroot = nx.NXroot(nx.NXdata(kwargs['img']))
    else:
        nxroot = nx.load(path, mode='rw')


    if not hasattr(nxroot.data, 'rawfile'):
        nxroot.data.rawfile = kwargs['rawpath']
    if not hasattr(nxroot.data, 'thumb'):
        nxroot.data.thumbnail = kwargs['thumb']
    # TODO: merge variation
    if not hasattr(nxroot.data, 'variation'):
        nxroot.data.variation = kwargs['variation'].items()

    if newfile:
        writenexus(nxroot, kwargs['path'])


def writenexus(nexroot, path):
    try:
        nexroot.save(path)
    except IOError:
        msg.logMessage('IOError: Check that you have write permissions.',msg.ERROR)


@debugtools.timeit
def thumbnail(img, factor=5):
    """
    Generate a thumbnail from an image
    """

    shape = img.shape[::-1]
    img = Image.fromarray(img)
    img.thumbnail(np.divide(shape, factor))
    # print 'thumb:', shape, np.array(img).shape
    return np.array(img)


import StringIO


@debugtools.timeit
def jpeg(img):
    buffer = StringIO.StringIO()
    pilImage = Image.fromarray(img)
    pilImage.save(buffer, "JPEG", quality=85)
    msg.logMessage(('JPEG buffer size (bytes):', buffer.len),msg.DEBUG)
    return buffer


def blockshaped(arr, factor):
    firstslice = np.array_split(arr, arr.shape[0] // factor)
    secondslice = map(lambda x: np.array_split(x, arr.shape[1] // factor, axis=1), firstslice)
    return np.array(secondslice)


def writeimage(image, path, mask=None, headers=None, suffix='',ext=None, dialog=False):
    if dialog:
        filename,ok = dialogs.savedatadialog(guesspath=path,caption="Save data to "+ext)
        if filename and ok:
            writeimage(image, filename, headers)
            if mask is not None:
                maskname = ''.join(os.path.splitext(filename)[:-1]) + "_mask" + os.path.splitext(filename)[-1]
                writeimage(mask, maskname, headers)

    if headers is None: headers = dict()
    if ext is None: ext = os.path.splitext(path)[-1]

    path = ''.join(os.path.splitext(path)[:-1]) + suffix + ext
    if notexitsoroverwrite(path):
        if ext.lower() == '.edf':
            fabimg = edfimage.edfimage(np.rot90(image), header=headers)
            fabimg.write(path)
        elif ext.lower() == '.tif':
            fabimg = tifimage.tifimage(np.rot90((image.astype(float)/image.max()*2**16).astype(np.int16)), header=headers)
            fabimg.write(path)
        elif ext.lower() == '.png':
            raise NotImplementedError
        elif ext.lower() == '.jpg':
            scipy.misc.imsave(path,np.rot90(image))

    else:
        return False
    return True


def writearray(data, path, headers=None, suffix=''):
    if headers is None: headers = dict()
    ext = '.csv'
    path = ''.join(os.path.splitext(path)[:-1]) + suffix + ext

    if notexitsoroverwrite(path):
        np.savetxt(path, np.array(data), header=headers)
        return True
    else:
        return False


def notexitsoroverwrite(path):
    if os.path.isfile(path):
        if not dialogs.checkoverwrite(path): return False
    return True


# -*- coding: utf-8 -*-


__author__ = "Luis Barroso-Luque, Holden Parks"
__copyright__ = "Copyright 2016, CAMERA, LBL, ALS"
__credits__ = ["Ronald J Pandolfi", "Dinesh Kumar", "Singanallur Venkatakrishnan", "Luis Luque",
               "Holden Parks", "Alexander Hexemer"]
__license__ = ""
__version__ = "1.2.1"
__maintainer__ = "Ronald J Pandolfi"
__email__ = "ronpandolfi@lbl.gov"
__status__ = "Beta"


import inspect
import time
import os
from collections import OrderedDict
from copy import deepcopy
from modpkgs import yamlmod
import numpy as np
from PySide import QtCore, QtGui
from . import config
from . import reconpkg
import functionwidgets as fc
from xicam.widgets import featurewidgets as fw
import dxchange
import tomopy


class FunctionManager(fw.FeatureManager):
    """
    Subclass of FeatureManager used to manage tomography workflow/pipeline FunctionWidgets

    Attributes
    ----------
    cor_offest : function/lambda
        function to correct for an offset in the COR location. As when padding the input array
    corr_scale : function/lambda
        function to correct for a scaling in the COR location. As when subsampling the input array
    recon_function : FunctionWidget
        FunctionWidget representing the Reconstruction Function in worflow pipeline
    slice_dir : dict
        Dictionary containing the axis that the function 'acts on'. Used for chunking and processing in reconstructions
    cor_func : functionwidgets.FunctionWidget
        FunctionWidget representing current COR detection widget in pipeline
    cor_widget : functionwidgets.CORSelectionWidget
        Widget for selecting method of COR detection (automatic/manual) and which function is used (in automatic case)
    Parameters
    ----------
    list_layout : QtGui.QLayout
        Layout to display the list of FunctionWidgets
    form_layout : QtGui.QLayout
        Layout to display the FunctionWidgets form (pyqtgraph.Parameter)
    function_widgets : list of FunctionWidgets, optional
        List with functionwidgets for initialization
    blank_form : QtGui.QWidget, optional
        Widget to display in form_layout when not FunctionWidget is selected

    Signals
    -------
    sigTestRange(str, object, dict)
        Emitted when request for TestParameterRange is submitted
    sigPipelineChanged()
        Emitted when the pipeline changes or the reconstruction function is changed
    sigCORDetectChanged(boo)
        Emitted when user toggles the COR detection method (between either automatic or manual detection)
    sigFuncAdded()
        Emitted when function is added to pipeline
    sigNormFuncAdded(QtGui.QWidget)
        Emitted when normalization function is added to pipeline
    """

    sigTestRange = QtCore.Signal(str, object, dict)
    sigPipelineChanged = QtCore.Signal()
    sigCORDetectChanged = QtCore.Signal(bool)
    sigFuncAdded = QtCore.Signal()
    sigNormFuncAdded = QtCore.Signal(QtGui.QWidget)

    center_func_slc = {'Phase Correlation': (0, -1)}  # slice parameters for center functions
    mask_functions = ['F3D Mask Filter', 'F3D MM Erosion', 'F3D MM Dilation', 'F3D MM Opening', 'F3D MM Closing']
    slice_dir = {
        'Remove Outliers': 'proj',
        'Remove NAN': 'proj',
        '1D Remove Outliers': 'sino',
        'Nearest Flats': 'sino',
        'Regular': 'both',
        'Background Intensity': 'proj',
        'ROI': 'both',
        'Beam Hardening': 'both',
        'Negative Logarithm': 'both',
        'Fourier-Wavelet': 'sino',
        'Titarenko': 'sino',
        'Smoothing Filter': 'sino',
        'Sino 360 to 180': 'sino',
        'correcttilt': 'proj',
        'Phase Retrieval': 'proj',
        'Gridrec': 'sino', 'FBP': 'sino', 'ART': 'sino', 'BART': 'sino', 'MLEM': 'sino', 'OSEm': 'sino',
        'OSPML Hybrid': 'sino', 'OSPML Quad': 'sino', 'PML Hybrid': 'sino', 'PML Quad': 'sino', 'SIRT': 'sino',
        'BP': 'sino', 'FBP_astra': 'sino', 'SIRT_astra': 'sino', 'SART': 'sino', 'ART_astra': 'sino', 'CGLS': 'sino', 'BP_CUDA': 'sino',
        'FBP_CUDA': 'sino', 'SIRT_CUDA': 'sino', 'SART_CUDA': 'sino', 'CGLS_CUDA': 'sino', 'Gridrec Tomocam': 'sino',
        'SIRT TomoCam': 'sino', 'MBIR TomoCam': 'sino',
        'Polar Mean Filter': 'sino',
        'F3D Bilateral Filter': 'both',
        'Convert Data Type': 'both',
        'Tiff Stack': 'both',
        'Write HDF5': 'both',
        'Padding': 'sino',
        'Crop': 'sino',
        'Slice': 'sino',
        'Downsample': 'both',
        'Upsample': 'both',
        'Circular Mask': 'sino',
        'Add': 'both', 'Subtract': 'both', 'Multiply': 'both', 'Divide': 'both', 'Maximum': 'both'
    }


    def __init__(self, list_layout, form_layout, function_widgets=None, blank_form=None):
        super(FunctionManager, self).__init__(list_layout, form_layout, feature_widgets=function_widgets,
                                              blank_form=blank_form)
        #TODO: add attribute to keep track of function's order in pipeline
        self.cor_offset = lambda x: x  # dummy
        self.cor_scale = lambda x: x  # dummy
        self.recon_function = None



    # TODO fix this astra check raise error if package not available
    def addFunction(self, function, subfunction, package):
        """
        Adds a Function to the workflow pipeline

        Parameters
        ----------
        function : str
            generic name of function
        subfunction : str
            specific name of function under the generic name category
        package : python package
            package where function is defined
        """
        if function == 'Reconstruction':
            if 'astra' in reconpkg.packages and package == reconpkg.packages['astra']:
                func_widget = fc.AstraReconFuncWidget(function, subfunction, package)
            elif 'mbir' in reconpkg.packages and package == reconpkg.packages['mbir']:
                func_widget = fc.TomoCamReconFuncWidget(function, subfunction, package)
            else:
                func_widget = fc.TomoPyReconFunctionWidget(function, subfunction, package)
            self.recon_function = func_widget
            func_widget.input_functions['center'].previewButton.clicked.connect(self.CORChoiceUpdated)
            self.sigPipelineChanged.emit()
        elif function == 'Normalization':
            func_widget = fc.NormalizeFunctionWidget(function, subfunction, package)
            self.sigNormFuncAdded.emit(func_widget)
        elif function == 'Filter' and subfunction in self.mask_functions:
            func_widget = fc.MaskFunctionWidget(function, subfunction, package)
        elif function == 'Reader':
            func_widget = fc.ReadFunctionWidget(function, subfunction, package)
        elif function == 'Write':
            func_widget = fc.WriteFunctionWidget(function, subfunction, package)
        else:
            func_widget = fc.FunctionWidget(function, subfunction, package)
        func_widget.sigTestRange.connect(self.testParameterRange)
        self.addFeature(func_widget)

        if function == 'Padding' or function  == 'Crop':
            self.connectCropPad()

        self.sigFuncAdded.emit()
        return func_widget


    def setFunctionValues(self, funcwidget, params, setDefault=False):
        """
        Sets a functionwidget's values based parameters in 'params'
        """
        for param, value in params.iteritems():
            try:
                funcwidget.params.child(param).setValue(value)
            except Exception:
                pass
            if setDefault:
                funcwidget.params.child(param).setDefault(value)


    def connectCropPad(self):
        """
        Connects 'Crop' and 'Pad' functions in pipeline, if they are both present, so that # of pixels added by
        padding matches the number removed by crop if either parameter is changed by the user
        """
        names = [func.name for func in self.features]
        if 'Padding' in names and 'Crop' in names:
            for feature in self.features:
                if feature.func_name == 'Padding': pad = feature
                elif feature. func_name == 'Crop': crop = feature

            pad.params.child('npad').sigValueChanged.connect(lambda x: crop.params.child('p11').setValue(x.value()))
            pad.params.child('npad').sigValueChanged.connect(lambda x: crop.params.child('p12').setValue(x.value()))
            pad.params.child('npad').sigValueChanged.connect(lambda x: crop.params.child('p21').setValue(x.value()))
            pad.params.child('npad').sigValueChanged.connect(lambda x: crop.params.child('p22').setValue(x.value()))

            children = [crop.params.child('p11'), crop.params.child('p21'), crop.params.child('p12'), crop.params.child('p22')]
            for p in children:
                for p_other in [param for param in children if param != p]:
                    p.sigValueChanged.connect(lambda x=p: p_other.setValue(x.value()))
                p.sigValueChanged.connect(lambda x: pad.params.child('npad').setValue(x.value()))

    def addInputFunction(self, funcwidget, parameter, function, subfunction, package, **kwargs):
        """
        Adds an input function to the give function widget

        Parameters
        ----------
        funcwidget : FunctionWidget
            Widget to add subfunction to
        parameter : str
            Parameter name that will be overriden by return value of input function
        function : str
            generic name of function
        subfunction : str
            specific name of function under the generic name category
        package : python package
            package where function is defined
        kwargs
            Additional keyword arguments
        """
        try:
            ipf_widget = fc.FunctionWidget(function, subfunction, package, **kwargs)
            funcwidget.addInputFunction(parameter, ipf_widget)
        except AttributeError:
            ipf_widget = funcwidget.input_functions[parameter]
        self.sigFuncAdded.emit()
        return ipf_widget

    def updateCORChoice(self, boolean):
        """
        Enables/disables the COR detection function in pipeline based on 'boolean' parameter, which is emitted when
        the user toggles the same option in the COR widget

        Parameters
        ----------
        boolean: bool
            'True' if automatic COR detection is enabled, 'False' otherwise
        """

        for feature in self.features:
            if 'center' in feature.input_functions:
                feature.input_functions['center'].enabled = boolean

    def updateCORFunc(self, func, widget):
        """
        Updates the COR detection function in the pipeline based on which function is selected in the COR widget

        Parameters
        ----------
        func : str
            Name of COR detection function chosen
        widget : FunctionWidget
            New COR detection FunctionWidget, located in COR widget
        """
        for feature in self.features:
            if 'center' in feature.input_functions:
                feature.removeInputFunction('center')
                feature.addCenterDetectFunction("Center Detection", func, package=reconpkg.packages['tomopy'])
                feature.input_functions['center'].previewButton.clicked.connect(self.CORChoiceUpdated)
                self.cor_func = feature.input_functions['center']
                self.cor_widget = widget
                for child in widget.params.children():
                    child.sigValueChanged.connect(self.updateCORPipeline)
                for child in self.cor_func.params.children():
                    child.sigValueChanged.connect(self.updateCORWidget)


    def updateCORPipeline(self, param):
        """
        Slot for connecting COR widget and pipeline COR functions - emitted when parameter in COR widget is changed,
        so that it can be changed in function pipeline as well

        Parameters
        ----------
        param : pyqtgraph.Parameter
            Parameter changed by user in COR widget
        """
        child = self.cor_func.params.child(param.name())
        child.setValue(param.value())

    # slot for connecting cor widget and pipeline cor
    def updateCORWidget(self, param):
        """
        Slot for connecting COR widget and pipeline COr functions - emitted when function in COR widget is changed, so
        it can be changed in pipeline can be changed
        """

        child = self.cor_widget.params.child(param.name())
        child.setValue(param.value())


    def CORChoiceUpdated(self):
        """
        Slot to receive signal when COR detection method is changed (auto/manual). Emits signal to reflect this
        in function pipeline
        """

        for feature in self.features:
            if 'center' in feature.input_functions:
                self.sigCORDetectChanged.emit(feature.input_functions['center'].enabled)


    def updateParameters(self):
        """
        Updates all parameters for the current function list
        """
        for function in self.features:
            function.updateParamsDict()

    def adjustReaderParams(self, roi_widget):
        """
        Adjusts reader functionwidget parameter based on dimensions specified by roi_widget parameter

        Parameters
        ----------
        roi_widget: pyqtgraph.ROI
            ROI to select dimensions for reader functionwidget
        """
        for feature in self.features:
            if 'Reader' in feature.name:
                pos = roi_widget.pos()
                size = roi_widget.size()
                x1 = int(pos[0]) if pos[0]>0 else 0
                x2 = int(pos[0]+size[0]) if pos[0]+size[0]<feature.params.child('end_width').defaultValue() else \
                        feature.params.child('end_width').defaultValue()
                y1 = int(pos[1]) if pos[1]>0 else 0
                y2 = int(pos[1]+size[1]) if pos[1]+size[1]<feature.params.child('end_sinogram').defaultValue() else \
                        feature.params.child('end_sinogram').defaultValue()

                feature.params.child('start_width').setValue(x1)
                feature.params.child('end_width').setValue(x2)
                feature.params.child('start_sinogram').setValue(y1)
                feature.params.child('end_sinogram').setValue(y2)


    def lockParams(self, boolean):
        """
        Locks all parameters for the current function list
        """
        for func in self.features:
            func.allReadOnly(boolean)

    def resetCenterCorrection(self):
        """
        Resets the center correction functions to dummy lambdas
        """
        self.cor_offset = lambda x: x  # dummy
        self.cor_scale = lambda x: x  # dummy

    def setCenterCorrection(self, name, param_dict):
        """
        Sets the center correction lambda's according to the effect of function given to the input array

        Parameters
        ----------
        name : str
            Name of function that has an effect on the COR value
        param_dict : dict
            Parameter dictionary of the function give
        """

        if 'Reader' in name:
            for feature in self.features:
                if 'Reader' in feature.name:
                    n = int(feature.params.child('start_width').value())
                    dummy = self.cor_offset
                    self.cor_offset = lambda x: dummy(x) - n
        elif 'Padding' in name and param_dict['axis'] == 2:
            n = param_dict['npad']
            dummy = self.cor_offset
            self.cor_offset = lambda x: dummy(x) + n
        elif 'Downsample' in name and param_dict['axis'] == 2:
            s = param_dict['level']
            self.cor_scale = lambda x: x / 2 ** s
        elif 'Upsample' in name and param_dict['axis'] == 2:
            s = param_dict['level']
            self.cor_scale = lambda x: x * 2 ** s

    def saveState(self, datawidget):
        """
        Parameters
        ----------

        datawidget : QWidget
            QWidget (usually in the form of a TomoViewer) that holds data

        Returns
        -------
        d : OrderedDict
            Dictionary of function names and corresponding partials
        theta : list
            array of angle values at which projection data was taken
        center : float
            Dataset center of rotation
        dict
            Dictionary used to write python file representing function pipeline
        dict
            Dictionary used to write yaml file representing function pipeline
        """
        self.lockParams(True)

        # extract function pipeline
        d = OrderedDict(); theta = []
        for function in self.features:
            if not function.enabled or 'Reader' in function.name:
                continue
            fpartial = function.partial
            # set keywords that will be adjusted later by input functions or users
            for arg in inspect.getargspec(function._function)[0]:
                if arg not in fpartial.keywords.iterkeys() or arg in 'center':
                    fpartial.keywords[arg] = '{}'.format(arg)

            # get rid of degenerate keyword arguments
            if 'arr' in fpartial.keywords and 'tomo' in fpartial.keywords:
                fpartial.keywords['tomo'] = fpartial.keywords['arr']
                fpartial.keywords.pop('arr', None)

            # special cases for the 'write' function
            if 'start' in fpartial.keywords:
                fpartial.keywords['start'] = 'start'
            if 'Write' in function.name:
                fpartial.keywords.pop('parent folder', None)
                fpartial.keywords.pop('folder name', None)
                fpartial.keywords.pop('file name', None)

            d[function.name] = fpartial

            if 'Reconstruction' in function.name: # could be bad, depending on if other operations need theta/center
                for param,ipf in function.input_functions.iteritems():
                    if not ipf.enabled:
                        if 'center' in param:
                            center = function.partial.keywords['center']
                        continue
                    # extract center value
                    if 'center' in param:
                        # this portion is taken from old updateFunctionPartial code
                        args = []
                        if ipf.subfunc_name in FunctionManager.center_func_slc:
                            map(args.append, map(datawidget.data.fabimage.__getitem__,
                                                         FunctionManager.center_func_slc[ipf.subfunc_name]))
                        else:
                            args.append(datawidget.getsino())
                        if ipf.subfunc_name == 'Nelder-Mead':
                            args.append(function.input_functions['theta'].partial())
                            # ipf.partial.keywords['theta'] = function.input_functions['theta'].partial()
                        center = ipf.partial(*args)

                    # extract theta values
                    if 'theta' in param:
                        theta = ipf.partial()
        self.lockParams(False)

        return [d, theta, center, config.extract_pipeline_dict(self.features),
                config.extract_runnable_dict(self.features)]

    def loadDataDictionary(self, datawidget, theta, center, slc = None):
        """
        Method to load dictionary of data relevant to a reconstruction

        Parameters
        ----------
        datawidget : QWidget
            QWidget (usually in the form of a TomoViewer) that holds data
        theta : array
            Array of values which represent angles at which tomography data was taken
        center : float
            Center of rotation of data
        slc : Slice
            Slice object to extract relevant portions of sinogram/flats/darks data

        Returns
        -------
        data_dict : dict
            Dictionary of data relevant to reconstruction
        """

        data_dict = OrderedDict()

        data_dict['tomo'] = datawidget.getsino(slc=slc)
        if datawidget.data.flats is not None:
            data_dict['flats'] = datawidget.getflats(slc=slc)
        if datawidget.data.darks is not None:
            data_dict['dark'] = datawidget.getdarks(slc=slc)

        # this ensures that if slc has '1' as its middle dimension, the array will be 3D
        # this is only necessary for slice previews, so we can assume that the middle axis
        # will be the dimension in question
        for key in ['tomo', 'flats', 'dark']:
            if len(data_dict[key].shape) < 3:
                data_dict[key] = data_dict[key][:, np.newaxis, :]


        try:
            if slc is not None and slc[0].start is not None:
                slc_ = (slice(slc[0].start, datawidget.data.shape[0] - 1, slc[0].step) if slc[0].stop is None
                        else slc[0])
                flat_loc = map_loc(slc_, datawidget.data.fabimage.flatindices())
            else:
                flat_loc = datawidget.data.fabimage.flatindices()
            data_dict['flat_loc'] = flat_loc
        except TypeError: pass
        except AttributeError: pass

        data_dict['theta'] = theta
        data_dict['center'] = center

        return data_dict



    def updatePartial(self, function, name, data_dict, param_dict, slc=None):
        """
        Updates the given function partial's keywords - the functools.partial object is the first element of func_tuple

        Parameters
        ----------

        func_tuple : functools.partial
            functools.partial of a function in the processing pipeline
        name :str
            Name of function being updated
        data_dict : dict
            Dictionary which contains all information relevant to a reconstruction - its elements are loaded into the
            functools.partial's keyword arguments
        param_dict : dict
            Copy of the original keyword arguments of the functools.partial, for reference

        Return
        ------

        function : functools.partial
            Function partial with all keyword arguments loaded
        write : str
            Name of the array function will act on - this allows users to specify which data a function will act on
            in the pipeline
        """

        write = 'tomo'

        for key, val in param_dict.iteritems():
            if val in data_dict.iterkeys():
                if 'arr' in key or 'tomo' in key:
                    write = val
                function.keywords[key] = data_dict[val]

        if 'Reconstruction' in name:
            slc_offset = 0 if (not slc or not slc[2].start) else slc[2].start
            function.keywords['center'] = self.cor_offset(self.cor_scale(function.keywords['center'])) - slc_offset
            self.resetCenterCorrection()
        elif name == 'Sino 360 to 180':
            # since this parameter needs the COR (which can change), loading must be done here
            # instead of done earlier
            tomo = data_dict['tomo']
            if tomo.shape[0]%2>0:
                function.keywords['overlap'] = int(np.round((tomo.shape[2] - data_dict['center'] -0.5))*2)
            else:
                function.keywords['overlap'] = int(np.round((tomo.shape[2] - data_dict['center']))*2)
        elif name == 'Write (Tiff Stack)':
            function.keywords['start'] = data_dict['y']*data_dict['num_sino_per_chunk'] + data_dict['sino'][0]
        return function, write

    def updateDataDict(self, data_dict, function, name):
        """
        Updates the data dictionary (which holds reconstruction data and other data used in reconstructin process)
        when functions are used which affect this data. Processing is done in place.

        Parameters
        ----------
        data_dict: dict
            Dictionary which contains all elements relevant to a reconstruction
        function: functools.partial
            functools.partial of a function in a pipeline
        name: str
            name of the function that has just been performed
        """

        # used after function has been performed
        # set center of rotation after changing data dimensions with these functions
        if name in ('Padding', 'Downsample', 'Upsample'):
            self.setCenterCorrection(name, function.keywords)

        # 360 to 180 changes many of these values, so we store them with a 'keepvalues' key in case they are
        # needed later
        if 'Sino 360 to 180' in name:
            data_dict['keepvalues'] = [data_dict['proj'], data_dict['num_proj_per_chunk'], data_dict['numprojchunks'],
                                       data_dict['numprojused'], data_dict['width']]
            proj = data_dict['proj']
            numangles = int(proj[1]-proj[0]) / 2
            data_dict['proj'] = (0, numangles - 1, proj[2])
            data_dict['num_proj_per_chunk'] = np.minimum(data_dict['num_proj_per_chunk'],
                                                         data_dict['proj'][1] - data_dict['proj'][0])
            data_dict['numprojchunks'] = (data_dict['proj'][1]-data_dict['proj'][0]-1) // data_dict['num_proj_per_chunk'] + 1
            data_dict['numprojused'] = (data_dict['proj'][1] - data_dict['proj'][0]) // data_dict['proj'][2]
            data_dict['width'] = (data_dict['width'][0], data_dict['width'][1] + data_dict['tomo'].shape[2],
                                  data_dict['width'][2])
            min_angle = data_dict['theta'][-1]; max_angle = data_dict['theta'][0]
            data_dict['theta'] = tomopy.angles(numangles, np.rad2deg(max_angle), np.rad2deg(min_angle))

    def loadPreviewData(self, datawidget, dims=None, ncore=None, skip_names=['Write', 'Reader'],
                        fixed_func=None, prange=None):
        """
        Create the function stack and summary dictionary used for running slice previews and 3D previews

        Parameters
        ----------

        datawidget
            Class holding the input dataset
        slc slice
            Slice object to extract tomography/flat/dark data when appropriate
        ncore : int
            number of cores to set the appropriate functions to run on
        skip_names : list of str, optional
            Names of functions to skip when running but still add to the dict representing the pipeline to run.
            Currently only the Writing functions are skipped as writing is not necessary in previews.
        fixed_func : type class
            A dynamic class with only the necessary attributes to be run in a workflow pipeline. This is used for
            parameter range tests to create the class with the parameter to be run and send it to a background thread.
            See testParameterRange for more details

        Returns
        -------
        partial_stack : list of partials:
            List with function partials needed to run preview
        self.stack_dict : dict
            Dictionary summarizing functions and parameters representing the pipeline (used for the list of partials)
        data_dict : dict
            Dictionary of data necessary to run a reconstruction preview
        """

        stack_dict = OrderedDict()
        self.lockParams(True)

        func_dict, theta, center, yaml_pipe, python_pipe = self.saveState(datawidget)
        keys = [key.split(' (')[0] for key in list(func_dict.keys())]
        for name in skip_names:
            if name in list(keys):
                ind = keys.index(name)
                func_dict.pop(list(func_dict.keys())[ind])

        # load data dictionary
        proj, sino, width, proj_p_chunk, sino_p_chunk, cpu_count = dims
        slc = (slice(*proj), slice(*sino), slice(*width))
        data_dict = self.loadDataDictionary(datawidget, theta, center, slc = slc)

        count = 1
        for func in self.features:
            name = func.name
            func_name = str(count) + ". " + func.func_name
            if not func.enabled:
                continue
            elif func.func_name in skip_names:
                stack_dict[func_name] = {func.subfunc_name: deepcopy(func.exposed_param_dict)}
                count += 1
                continue
            elif fixed_func is not None and func.func_name == fixed_func.func_name:
                func = fixed_func
                for key, val in func.exposed_param_dict.iteritems():
                    if key in 'center':
                        center = func.exposed_param_dict[key]
                    elif key in func_dict[name].keywords.iterkeys() and key not in 'center':
                        func_dict[name].keywords[key] = val
            stack_dict[func_name] = {func.subfunc_name: deepcopy(func.exposed_param_dict)}
            count += 1

            for param, ipf in func.input_functions.iteritems():
                if ipf.enabled:
                    if 'Input Functions' not in stack_dict[func_name][func.subfunc_name]:
                        stack_dict[func_name][func.subfunc_name]['Input Functions'] = {}
                    ipf_dict = {param: {ipf.func_name: {ipf.subfunc_name: ipf.exposed_param_dict}}}
                    stack_dict[func_name][func.subfunc_name]['Input Functions'].update(ipf_dict)

                    # update input function keywords in slice preview table
                    if param in stack_dict[func_name][func.subfunc_name]:
                        stack_dict[func_name][func.subfunc_name][param] = data_dict[param]

        self.lockParams(False)
        return datawidget, func_dict, theta, center, stack_dict, prange, dims



    def reconGenerator(self, datawidget, func_dict, theta, center, yaml_pipe,
                       run_dict, dims, ncore = 1, reconType='Full'):

        """
        Generator for running full reconstruction. Yields messages representing the status of reconstruction or the
        tomography data itself. This function chunks the data into portions of either sinograms of projections. The
        chunking is necessary because datasets are often too large to fit entirely into memory. The alternating between
        sinograms and projections is necessary because some functions need *all*  of the dataset's sinograms/projections
        to work properly (for example, phase retrieval needs all of the dataset's sinograms, so it chunks the dataset's
        projections).

        See https://bitbucket.org/berkeleylab/als-microct-python for script on which this function is based

        Parameters
        ----------
        datawidget : QWidget
            QWidget (usually in the form of a TomoViewer) that holds data
        run_state : list of four elements
            * list of functools.partial which represent the function pipeline
            * array of 'theta' values which represent the angles at which tomography data was taken
            * the center of rotation of the data
            * a string of function names and parameters to be later written into a yaml file
        proj : tuple of int
            Projection range indices (start, end, step)
        sino : tuple of int
            Sinogram range indices (start, end, step)
        sino_p_chunk : int
            Number of sinograms per chunk
        proj_p_chunk : int
            Number of projections per chunk
        ncore : int
            Number of cores to run functions
        pipeline_dict: dictionary
            Dictionary of parameters referenced during reconstruction

        Yields
        -------
        str
            Message of current status of function
        d['tomo'] : np.ndarray
            Tomography data after all processing is complete
        """

        start_time = time.time()
        proj, sino, width, sino_p_chunk, proj_p_chunk, cpu_count = dims
        write_start = sino[0]
        path = datawidget.path
        preview_slices = []

        num_proj_per_chunk = np.minimum(proj_p_chunk, proj[1] - proj[0])
        numprojchunks = (proj[1] - proj[0] - 1)//num_proj_per_chunk+1
        num_sino_per_chunk = np.minimum(sino_p_chunk, sino[1] - sino[0])
        numsinochunks = (sino[1]-sino[0]-1)//num_sino_per_chunk+1
        numprojused = (proj[1] - proj[0])//proj[2]
        numsinoused = (sino[1] - sino[0])//sino[2]

        # set up dictionary of function keywords
        params_dict = OrderedDict()
        for name, fpartial in list(func_dict.items()):
            params_dict['{}'.format(name)] = dict(fpartial.keywords)


        if reconType == 'Full':
            # get save names for pipeline yaml/runnable files
            dir = ""
            for name, fpartial in list(func_dict.items()):
                if 'fname' in fpartial.keywords:
                    fname = fpartial.keywords['fname']
                    for item in fname.split('/')[:-1]:
                        dir += item + '/'
                    yml_file = fname + '.yml'
                    python_file = fname + '.py'

            # make project directory if it isn't made already
            if not os.path.exists(dir):
                os.makedirs(dir)

            # save function pipeline as runnable
            yield "Start {} at: ".format(path) + time.strftime("%a, %d %b %Y %H:%M:%S +0000", time.localtime())

            # for now, do not write out python script on reconstruction
            # runnable = self.extractPipelineRunnable(run_dict, theta, params_dict, proj, sino, sino_p_chunk, width, path, ncore)
            # try:
            #     with open(python_file, 'w') as py:
            #         py.write(runnable)
            # except NameError or IOError:
            #     yield "Error: pipeline python script not written - path could not be found"

            # save yaml in reconstruction folder
            for key in yaml_pipe.iterkeys(): # special case for 'center' param
                if 'Recon' in key:
                    for subfunc in yaml_pipe[key].iterkeys():
                        if 'Parameters' in yaml_pipe[key][subfunc].iterkeys():
                            yaml_pipe[key][subfunc]['Parameters']['center'] = float(center)


            try:
                with open(yml_file, 'w') as yml:
                    yamlmod.ordered_dump(yaml_pipe, yml)
            except NameError or IOError:
                yield "Error: function pipeline yaml not written - path could not be found"

        # determine which direction to slice in first
        for name in list(func_dict.keys()):
            subfunc = name.split('(')[-1].split(')')[0]
            if self.slice_dir[subfunc] != 'both':
                axis = self.slice_dir[subfunc]
                break

        curfunc = 0
        curtemp = 0
        tempfilenames = [os.path.join(os.path.dirname(path), 'tmp0.h5'),
                         os.path.join(os.path.dirname(path), 'tmp1.h5')]

        d = {'num_proj_per_chunk': num_proj_per_chunk, 'num_sino_per_chunk':
                    num_sino_per_chunk, 'numprojchunks': numprojchunks, 'numsinochunks': numsinochunks,
                    'numprojused': numprojused, 'numsinoused': numsinoused,
                    'proj': proj, 'sino': sino, 'width': width, 'keepvalues': None}

        while True:
            if axis=='proj':
                niter = d['numprojchunks']
            else:
                niter = d['numsinochunks']
            for y in range(niter):
                # reset COR change for downsampling in 3D recon previews
                if axis != 'proj' and reconType == '3D':
                    self.cor_scale = lambda x: x // width[2]
                d['y'] = y
                yield "{} chunk {} of {}".format(axis, y+1, niter)
                if curfunc==0:
                    if axis=='proj':
                        slc = (slice(y*num_proj_per_chunk+proj[0], np.minimum((y+1)*num_proj_per_chunk+proj[0], proj[1]),
                                     proj[2]), slice(*sino), slice(*width))
                    else:
                        slc = (slice(*proj), slice(y*num_sino_per_chunk+sino[0], np.minimum((y+1)*num_sino_per_chunk
                                + sino[0], sino[1]), sino[2]), slice(*width))
                    tmp_d = self.loadDataDictionary(datawidget, theta, center,  slc=slc)
                    d.update(tmp_d)

                else:
                    if axis=='proj':
                        start, end = y * d['num_proj_per_chunk'], np.minimum((y + 1)*d['num_proj_per_chunk'], d['numprojused'])
                        tomo = dxchange.reader.read_hdf5(tempfilenames[curtemp], '/tmp/tmp',
                                slc=((start, end, 1), (0, d['numsinoused'], 1),
                                     (0, (d['width'][1]-d['width'][0])//d['width'][2], 1)))
                    else:
                        start, end = y * d['num_sino_per_chunk'], np.minimum((y + 1)*d['num_sino_per_chunk'], d['numsinoused'])
                        tomo = dxchange.reader.read_hdf5(tempfilenames[curtemp], '/tmp/tmp',
                                slc=((0, d['numprojused'], 1), (start, end, 1),
                                     (0, (d['width'][1]-d['width'][0])//d['width'][2], 1)))
                    d['tomo'] = tomo
                dofunc = curfunc
                d['start'] = write_start
                d['keepvalues'] = None
                while True:
                    func_name = func_dict.keys()[dofunc]
                    subfunc = func_name.split('(')[-1].split(')')[0]
                    newaxis = self.slice_dir[subfunc]
                    if newaxis != 'both' and newaxis != axis:
                        # we have to switch the axis, so flush to disk
                        if y ==0:
                            try:
                                os.remove(tempfilenames[1-curtemp])
                            except OSError:
                                pass
                        appendaxis = 1 if axis=='sino' else 0
                        dxchange.write_hdf5(d['tomo'], fname=tempfilenames[1-curtemp], gname='tmp', dname='tmp',
                                            overwrite=False, appendaxis=appendaxis)
                        break
                    yield "{}: ".format(func_name)
                    curtime = time.time()
                    function, write = self.updatePartial(func_dict[func_name], func_name, d,
                                                         params_dict[func_name], slc)
                    d[write] = function()
                    self.updateDataDict(d, func_dict[func_name], func_name)

                    yield 'Finished in {:2f} seconds\n'.format(time.time()-curtime)
                    dofunc += 1
                    if dofunc == len(func_dict):
                        preview_slices.append(d['tomo']) # append to list for previews
                        break
                if y < niter - 1 and d['keepvalues']:
                    d['proj'], d['num_proj_per_chunk'], d['numprojchunks'], d['numprojused'], d['width'] = d['keepvalues']
                if axis == 'sino':
                    write_start += num_sino_per_chunk
            curtemp = 1 - curtemp
            curfunc = dofunc
            if curfunc == len(func_dict):
                break
            axis = self.slice_dir[func_dict.keys()[curfunc].split('(')[-1].split(')')[0]]
        yield "cleaning up temp files"
        for tmpfile in tempfilenames:
            try:
                os.remove(tmpfile)
            except OSError:
                pass

        # for the reconstruction previews
        if reconType == '3D' or reconType == 'Slice':
            d['tomo'] = np.vstack(preview_slices)
            yield d['tomo']

        yield "End Time: " + time.strftime("%a, %d %b %Y %H:%M:%S +0000", time.localtime())
        yield "It took {:3f} s to process {}".format(time.time() - start_time, path)


    def testParameterRange(self, function, parameter, prange):
        """
        Used to set off parameter range tests. Emits sigTestRange with message and a fixed_func object representing the
        function who's parameter is to be changed

        Parameters
        ----------
        function : FunctionWidget
            Widget containing the parameter to be evaluated
        parameter : str
            Parameter name
        prange : tuple/list
            Range of parameters to be evaluated
        """
        self.updateParameters()
        if function.func_name in 'Reader':
            return
        if 'Reconstruction' in function.func_name and parameter == 'center':
            function.input_functions['center'].enabled = False
        for i in prange:
            function.param_dict[parameter] = i
            # Dynamic FixedFunc "dummed down" FuncWidget class. cool.
            fixed_func = type('FixedFunc', (), {'func_name': function.func_name, 'subfunc_name': function.subfunc_name,
                                                'missing_args': function.missing_args,
                                                'param_dict': function.param_dict,
                                                'exposed_param_dict': function.exposed_param_dict,
                                                'partial': function.partial,
                                                'input_functions': function.input_functions,
                                                '_function': function._function})
            self.sigTestRange.emit('Computing preview for {} parameter {}={}...'.format(function.name, parameter, i),
                                   fixed_func, {'function': function.func_name, parameter: prange})


    def setPipelineFromYAML(self, pipeline, setdefaults=False, config_dict=config.names):
        """
        Sets the managers function/feature list from a dictionary from a YAML file.

        Parameters
        ----------
        pipeline : dict
            Dict extracted from YAML file
        setdefaults : bool
            Set the given parameter values as defaults
        config_dict : dict, optional
            Dictionary with configuration specifications/function parameter details
        """

        self.removeAllFeatures()
        # Way too many for loops, oops... may want to restructure the yaml files
        for func, subfuncs in pipeline.iteritems():
            try:
                func = func.split(". ")[1]
            except IndexError:
                func = func
            for subfunc in subfuncs:
                funcWidget = self.addFunction(func, subfunc, package=reconpkg.packages[config_dict[subfunc][1]])
                if 'Enabled' in subfuncs[subfunc] and not subfuncs[subfunc]['Enabled']:
                    funcWidget.enabled = False
                if 'Parameters' in subfuncs[subfunc]:
                    self.setFunctionValues(funcWidget, subfuncs[subfunc]['Parameters'], setdefaults)
                if 'Input Functions' in subfuncs[subfunc]:
                    for param, ipfs in subfuncs[subfunc]['Input Functions'].iteritems():
                        for ipf, sipfs in ipfs.iteritems():
                            for sipf in sipfs:
                                if param in funcWidget.input_functions:
                                    ifwidget = funcWidget.input_functions[param]
                                else:
                                    ifwidget = self.addInputFunction(funcWidget, param, ipf, sipf,
                                                                     package=reconpkg.packages[config_dict[sipf][1]])
                                if 'Enabled' in sipfs[sipf] and not sipfs[sipf]['Enabled']:
                                    ifwidget.enabled = False
                                if 'Parameters' in sipfs[sipf]:
                                    self.setFunctionValues(ifwidget, sipfs[sipf]['Parameters'], setdefaults)
                                ifwidget.updateParamsDict()
                funcWidget.updateParamsDict()
        self.sigPipelineChanged.emit()

    def setPipelineFromDict(self, pipeline, config_dict=config.names):
        """
        Sets the managers function/feature list from a dictionary extracted from a summary dictionary as the ones
        displayed in previews

        Parameters
        ----------
        pipeline : dict
            Dict representing the workflow pipeline
        config_dict : dict, optional
            Dictionary with configuration specifications/function parameter details
        """

        self.removeAllFeatures()
        for func, subfuncs in pipeline.iteritems():
            try:
                func = func.split(". ")[1]
            except IndexError:
                func = func
            for subfunc in subfuncs:
                funcWidget = self.addFunction(func, subfunc, package=reconpkg.packages[config_dict[subfunc][1]])
                for param, value in subfuncs[subfunc].iteritems():
                    if param == 'Package':
                        continue
                    elif param == 'Input Functions':
                        for param, ipfs in value.iteritems():
                            for ipf, sipf in ipfs.iteritems():
                                ifwidget = self.addInputFunction(funcWidget, param, ipf, sipf.keys()[0],
                                                    package=reconpkg.packages[config_dict[sipf.keys()[0]][1]])
                                for p, v in sipf[sipf.keys()[0]].items():
                                    ifwidget.params.child(p).setValue(v)
                                ifwidget.updateParamsDict()
                    else:
                        funcWidget.params.child(param).setValue(value)
                    funcWidget.updateParamsDict()
        self.sigPipelineChanged.emit()

    def extractPipelineRunnable(self, runnable_pipe, center, params,
                                proj, sino, sino_p_chunk, width, path, ncore=1):
        """
        Saves the function pipeline as a runnable (Python) file. This needs more work.

        # TODO: make new backend pipeline compatible with python script
        """


        signature = "import time \nimport tomopy \nimport dxchange\nimport h5py\n" \
                    "import numpy as np\nimport numexpr as ne\n\n"

        # set up function pipeline
        func_dict = runnable_pipe['func']
        subfunc_dict = runnable_pipe['subfunc']

        signature += "def main():\n\n"
        signature += "\t# offset and scale factors in case of padding, upsampling, or downsampling\n"
        signature += "\tcor_offset = 0\n"
        signature += "\tcor_scale = 0\n\n"
        signature += "\tstart_time = time.time()\n\n"

        signature += "\n\tdata = dxchange.read_als_832h5('{}')\n".format(path)
        signature += "\ttomo=data[0]; flats=data[1]; dark=data[2]\n"
        signature += "\tmdata = read_als_832h5_metadata('{}')\n\n".format(path)
        signature += "\t# choose which projections and sinograms go into the reconstruction:\n"
        signature += "\tproj_start = {}; proj_end = {}; proj_step = {}\n".format(proj[0],proj[1],proj[2])
        signature += "\tsino_start = {}; sino_end = {}; sino_step = {}\n".format(sino[0],sino[1],sino[2])
        signature += "\twidth_start = {}; width_end = {}; width_step = 1\n".format(width[0], width[1])
        signature += "\tsino_p_chunk = {2} # chunk size of data during reconstruction\n\n".format(proj, sino, sino_p_chunk)
        signature += "\tproj = (proj_start, proj_end, proj_step)\n"
        signature += "\tsino = (sino_start, sino_end, sino_step)\n"
        signature += "\twidth = (width_start, width_end, width_step)\n\n"
        signature += "\twrite_start = sino[0]\n"
        signature += "\tnchunk = ((sino[1]-sino[0]) // sino[2] - 1) // sino_p_chunk +1\n"
        signature += "\ttotal_sino = (sino[1] - sino[0] - 1) // sino[2] + 1\n"
        signature += "\tif total_sino < sino_p_chunk:\n\t\tsino_p_chunk = total_sino\n\n"

        if 'center' not in subfunc_dict.iterkeys():
            signature += "\tcenter = {}\n".format(center)
        for key, val in subfunc_dict.iteritems():
            signature += "\t{} = {}\n".format(key, val)

        signature += "\n\n\t# MAIN LOOP FOR RECONSTRUCTION\n"
        signature += "\tfor i in range(nchunk):\n"
        signature += "\t\tstart, end = i * sino[2] * sino_p_chunk + sino[0], (i + 1) * sino[2] * sino_p_chunk + sino[0]\n"
        signature += "\t\tend = end if end < sino[1] else sino[1]\n\n"
        signature += "\t\tslc = (slice(*proj), slice(start,end,sino[2]), slice(*width))\n"

        signature += "\t\tdata_dict = loadDataDict(data, mdata, theta, center, slc)\n"
        signature += "\t\tdata_dict['start'] = write_start\n"
        signature += "\t\tshape = data_dict['tomo'].shape[1]\n\n"

        signature += "\t\t# the function pipeline: keywords used in each function are located in the\n"
        signature += "\t\t# 'params' assignment for each function\n\n"
        for func, param_dict in func_dict.iteritems():
            try:
                func = func.split(". ")[1]
            except IndexError:
                pass
            signature += "\t\t# function: {}\n".format(func)
            signature += "\t\tts = time.time()\n"
            signature += "\t\tprint 'Running {0} on slices {1} to {2} from a total of {3} slices'.format("
            signature += "'{}', start, end, total_sino)\n ".format('{}'.format(func))
            signature += "\t\tparams = {}\n".format(param_dict)
            signature += "\t\tkwargs, write, cor_offset, cor_scale = updateKeywords('{}', params, data_dict,".format(func)
            signature += " cor_offset, cor_scale, width)\n"
            signature += "\t\tdata_dict[write] = {}(**kwargs)\n".format(func)
            signature += "\t\tprint 'Finished in {:.3f} s'.format(time.time()-ts)\n"
            signature += "\t\tprint "" #white space \n\n"
        signature += "\t\twrite_start += shape\n\n"
        signature += "\tprint 'Reconstruction complete. Run time: {:.2f} s'.format(time.time()-start_time)\n"
        signature += "\tprint # white space\n\n"

        # rewrite functions used for processing
        signature += "# helper functions\n\n"
        signature += "def read_als_832h5_metadata(fname):\n"
        signature += "\t with h5py.File(fname, 'r') as f:\n\t\tg=_find_dataset_group(f)\n\t\treturn dict(g.attrs)\n\n"

        signature += "def _find_dataset_group(h5object):\n"
        signature += "\tkeys = h5object.keys()\n \tif len(keys)==1:\n"
        signature += "\t\tif isinstance(h5object[keys[0]], h5py.Group):\n"
        signature += "\t\t\tgroup_keys = h5object[keys[0]].keys()\n"
        signature += "\t\t\tif isinstance(h5object[keys[0]][group_keys[0]], h5py.Dataset):\n"
        signature += "\t\t\t\treturn h5object[keys[0]]\n"
        signature += "\t\t\telse:\n\t\t\t\treturn _find_dataset_group(h5object[keys[0]])\n"
        signature += "\t\telse:\n\t\t\traise Exception('Unable to find dataset group')\n"
        signature += "\telse:\n\t\traise Exception('Unable to find dataset group')\n\n"

        signature += "def flatindices(mdata):\n"
        signature += "\ti0 = int(mdata['i0cycle'])\n\tnproj = int(mdata['nangles'])\n"
        signature += "\tif i0 > 0:\n\t\tindices = list(range(0, nproj, i0))\n"
        signature += "\t\tif indices[-1] != nproj - 1:\n\t\t\tindices.append(nproj - 1)\n"
        signature += "\telif i0 == 0:\n\t\tindices = [0, nproj - 1]\n\treturn indices\n\n"

        signature += "# sets COR correction in case of padding, upsample, or downsample\n"
        signature += "def setCenterCorrection(name, param_dict, cor_offset, cor_scale):\n"
        signature += "\tif 'pad' in name and param_dict['axis'] == 2:\n"
        signature += "\t\tn = param_dict['npad']\n"
        signature += "\t\tcor_offset = n\n"
        signature += "\telif 'downsample' in name and param_dict['axis'] == 2:\n"
        signature += "\t\ts = param_dict['level']\n"
        signature += "\t\tcor_scale = -s\n"
        signature += "\telif 'upsample' in name and param_dict['axis'] == 2:\n"
        signature += "\t\ts = param_dict['level']\n"
        signature += "\t\tcor_scale = -s\n"
        signature += "\treturn cor_offset, cor_scale\n\n"
        signature += "def resetCenterCorrection(cor_offset, cor_scale):\n"
        signature += "\tcor_offset = 0\n\tcor_scale = 0\n"
        signature += "\treturn cor_offset, cor_scale\n\n"

        signature += "# performs COR correction\n"
        signature += "def correctCenter(center, cor_offset, cor_scale, width):\n"
        signature += "\tif cor_scale<0:\n\t\treturn float(int(center * 2 ** cor_scale)) + cor_offset - width[0]\n"
        signature += "\telse:\n\t\treturn (center * 2 ** cor_scale) + cor_offset - width[0]\n\n"


        signature += "def map_loc(slc,loc):\n\tstep = slc.step if slc.step is not None else 1\n"
        signature += "\tind = range(slc.start, slc.stop, step)\n\tloc = np.array(loc)\n\tlow, upp = ind[0], ind[-1]\n"
        signature += "\tbuff = (loc[-1] - loc[0]) / len(loc)\n\tmin_loc = low - buff\n\tmax_loc = upp + buff\n"
        signature += "\tloc = np.intersect1d(loc[loc > min_loc], loc[loc < max_loc])\n\tnew_upp = len(ind)\n"
        signature += "\tloc = (new_upp * (loc - low)) // (upp - low)\n\tif loc[0] < 0:\n\t\tloc[0] = 0\n"
        signature += "\treturn np.ndarray.tolist(loc)\n\n"

        # function for loading data dictionary
        signature += "def loadDataDict(data, mdata, theta,center,slc=None):\n\tdata_dict = {}\n"
        signature += "\tif slc is not None and slc[0].start is not None:\n"
        signature += "\t\tslc_ = slice(slc[0].start,data[0].shape[0],slc[0].step)\n"
        signature += "\t\tflat_loc = map_loc(slc_, flatindices(mdata))\n"
        signature += "\telse:\n\t\tflat_loc = flatindices(mdata)\n\n"
        signature += "\tdata_dict['tomo'] = data[0][slc]\n\tdata_dict['flats'] = data[1][slc]\n"
        signature += "\tdata_dict['dark'] = data[2][slc]\n\tdata_dict['flat_loc'] = flat_loc\n"
        signature += "\tdata_dict['theta'] = theta\n\tdata_dict['center'] = center\n\n"
        signature += "\treturn data_dict\n\n"

        signature += "def updateKeywords(function, param_dict, data_dict, cor_offset, cor_scale, width):\n"
        signature += "\twrite = 'tomo'\n"
        signature += "\tfor key, val in param_dict.iteritems():\n"
        signature += "\t\tif val in data_dict.iterkeys():\n"
        signature += "\t\t\tif 'arr' in key or 'tomo' in key:\n"
        signature += "\t\t\t\twrite = val\n"
        signature += "\t\t\tparam_dict[key] = data_dict[val]\n"
        signature += "\tfor item in ('pad', 'downsample', 'upsample'):\n"
        signature += "\t\tif item in function:\n "
        signature += "\t\t\tcor_offset, cor_scale = setCenterCorrection(function, param_dict, cor_offset, cor_scale)\n"
        signature += "\tif 'recon' in function:\n"
        signature += "\t\tparam_dict['center'] = correctCenter(param_dict['center'], cor_offset, cor_scale, width)\n"
        signature += "\t\tcor_offset, cor_scale = resetCenterCorrection(cor_offset, cor_scale)\n"
        signature += "\treturn param_dict, write, cor_offset, cor_scale\n\n\n"

        # write custom functions as functions in python file
        signature += "# the following three functions may be used in the reconstruction pipeline\n"
        signature += "def crop(arr, p11, p12, p21, p22, axis=0):\n"
        signature += "\tslc = []\n"
        signature += "\tpts = [p11, p12, p21, p22]\n"
        signature += "\tfor n in range(len(arr.shape)):\n"
        signature += "\t\tif n == axis:\n"
        signature += "\t\t\tslc.append(slice(None))\n"
        signature += "\t\telse:\n"
        signature += "\t\t\tslc.append(slice(pts.pop(0), -pts.pop(0)))\n"
        signature += "\treturn arr[slc]\n\n"

        signature += "def convert_data(arr, imin=None, imax=None, dtype='uint8', intcast='float32'):\n"
        signature += "\tDTYPE_RANGE = {'uint8': (0, 255), 'uint16': (0, 65535), 'int8': (-128, 127),"
        signature += "'int16': (-32768, 32767),'float32': (-1, 1),'float64': (-1, 1)}\n"
        signature += "\tallowed_dtypes = ('uint8', 'uint16', 'int8', 'int16', 'float32', 'float64')\n"
        signature += "\tif dtype not in allowed_dtypes:\n"
        signature += "\t\traise ValueError('dtype keyword {0} not in allowed keywords {1}'.format(dtype, allowed_dtypes))\n\n"

        signature += "\t# Determine range to cast values\n"
        signature += "\tminset = False\n"
        signature += "\tif imin is None:\n"
        signature += "\t\timin = np.min(arr)\n"
        signature += "\t\tminset = True\n"
        signature += "\tmaxset = False\n"
        signature += " \tif imax is None:\n"
        signature += "\t\timax = np.max(arr)\n"
        signature += "\t\tmaxset = True\n\n"

        signature += "\tnp_cast = getattr(np, str(arr.dtype))\n"
        signature += "\timin, imax = np_cast(imin), np_cast(imax)\n"

        signature += "\t# Determine range of new dtype\n"
        signature += "\tomin, omax = DTYPE_RANGE[dtype]\n"
        signature += "\tomin = 0 if imin >= 0 else omin\n"
        signature += "\tomin, omax = np_cast(omin), np_cast(omax)\n"

        signature += "\tif arr.dtype in [np.int8, np.int16, np.int32, np.int64, np.uint8, np.uint16,"
        signature += "np.uint32, np.uint64, np.bool_, np.int_, np.intc, np.intp]:\n"
        signature += "\t\tint_cast = getattr(np, str(intcast))\n"
        signature += "\t\tout = np.empty(arr.shape, dtype=int_cast)\n"
        signature += "\t\timin = int_cast(imin)\n"
        signature += "\t\timax = int_cast(imax)\n"
        signature += "\t\tdf = int_cast(imax) - int_cast(imin)\n"
        signature += "\telse:\n"
        signature += "\t\tout = np.empty(arr.shape, dtype=arr.dtype)\n"
        signature += "\t\tdf = imax - imin\n"
        signature += "\tif not minset:\n"
        signature += "\t\tif np.min(arr) < imin:\n"
        signature += "\t\t\tarr = ne.evaluate('where(arr < imin, imin, arr)', out=out)\n"
        signature += "\tif not maxset:\n"
        signature += "\t\tif np.max(arr) > imax:\n"
        signature += "\t\t\tarr = ne.evaluate('where(arr > imax, imax, arr)', out=out)\n"
        signature += "\tne.evaluate('(arr - imin) / df', truediv=True, out=out)\n"
        signature += "\tne.evaluate('out * (omax - omin) + omin', out=out)\n\n"

        signature += "\t# Cast data to specified type\n"
        signature += "\treturn out.astype(np.dtype(dtype), copy=False)\n\n"

        signature += """
def array_operation_add(arr, value=0):
    return ne.evaluate('arr + value')

def array_operation_sub(arr, value=0):
    return ne.evaluate('arr - value', truediv=True)

def array_operation_mult(arr, value=1):
    return ne.evaluate('arr * value')

def array_operation_div(arr, value=1):
    return ne.evaluate('arr / value')

def array_operation_max(arr, value=0):
    return np.maximum(arr, value)"""

        signature += "\n\nif __name__ == '__main__':\n\tmain()"

        return signature



def map_loc(slc, loc):
    """
    Does a linear mapping of the indices in loc from a range given by slc start and stop with step of one to a new
    range given by len(range(slc.start, slc.stop, slc.step))

    Parameters
    ----------
    slc : slice
    loc : list
        list of indices assumed to span from slc.start to slc.stop

    Returns
    -------
    list
        mapped indices to new range

    """

    step = slc.step if slc.step is not None else 1
    ind = range(slc.start, slc.stop, step)
    loc = np.array(loc)
    low, upp = ind[0], ind[-1]
    buff = (loc[-1] - loc[0]) / len(loc)
    min_loc = low - buff
    max_loc = upp + buff
    loc = np.intersect1d(loc[loc > min_loc], loc[loc < max_loc])
    new_upp = len(ind)
    loc = (new_upp * (loc - low)) // (upp - low)
    if loc[0] < 0:
        loc[0] = 0

    return np.ndarray.tolist(loc)

class TestRangeDialog(QtGui.QDialog):
    """
    Simple QDialog subclass with three spinBoxes to inter start, end, step for a range to test a particular function
    parameter
    """

    def __init__(self, dtype, prange, **opts):
        super(TestRangeDialog, self).__init__(**opts)
        SpinBox = QtGui.QSpinBox if dtype == 'int' else QtGui.QDoubleSpinBox
        self.gridLayout = QtGui.QGridLayout(self)
        self.spinBox = SpinBox(self)
        self.gridLayout.addWidget(self.spinBox, 1, 0, 1, 1)
        self.spinBox_2 = SpinBox(self)
        self.gridLayout.addWidget(self.spinBox_2, 1, 1, 1, 1)
        self.spinBox_3 = SpinBox(self)
        self.gridLayout.addWidget(self.spinBox_3, 1, 2, 1, 1)
        self.label = QtGui.QLabel(self)
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)
        self.label_2 = QtGui.QLabel(self)
        self.label_2.setAlignment(QtCore.Qt.AlignCenter)
        self.gridLayout.addWidget(self.label_2, 0, 1, 1, 1)
        self.label_3 = QtGui.QLabel(self)
        self.label_3.setAlignment(QtCore.Qt.AlignCenter)
        self.gridLayout.addWidget(self.label_3, 0, 2, 1, 1)
        self.buttonBox = QtGui.QDialogButtonBox(self)
        self.buttonBox.setOrientation(QtCore.Qt.Vertical)
        self.buttonBox.setStandardButtons(QtGui.QDialogButtonBox.Cancel | QtGui.QDialogButtonBox.Ok)
        self.gridLayout.addWidget(self.buttonBox, 0, 3, 2, 1)

        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        if prange is not (None, None, None):
            self.spinBox.setMaximum(9999)  # 3 * prange[2])
            self.spinBox_2.setMaximum(9999)  # 3 * prange[2])
            self.spinBox_3.setMaximum(9999)  # prange[2])

            self.spinBox.setValue(prange[0])
            self.spinBox_2.setValue(prange[1])
            self.spinBox_3.setValue(prange[2])

        self.setWindowTitle("Set parameter range")
        self.label.setText("Start")
        self.label_2.setText("End")
        self.label_3.setText("Step")

    def selectedRange(self):
        # return the end as selected end + step so that the range includes the end
        return np.arange(self.spinBox.value(), self.spinBox_2.value(), self.spinBox_3.value())

class TestListRangeDialog(QtGui.QDialog):
    """
    Simple QDialog subclass with comboBox and lineEdit to choose from a list of available function parameter keywords
    in order to test the different function parameters.
    """

    def __init__(self, options, **opts):
        super(TestListRangeDialog, self).__init__(**opts)
        self.gridLayout = QtGui.QGridLayout(self)
        self.comboBox = QtGui.QComboBox(self)
        self.gridLayout.addWidget(self.comboBox, 1, 0, 1, 1)
        self.lineEdit = QtGui.QLineEdit(self)
        self.lineEdit.setReadOnly(True)
        self.gridLayout.addWidget(self.lineEdit, 2, 0, 1, 1)
        self.buttonBox = QtGui.QDialogButtonBox(self)
        self.buttonBox.setOrientation(QtCore.Qt.Vertical)
        self.buttonBox.setStandardButtons(QtGui.QDialogButtonBox.Cancel | QtGui.QDialogButtonBox.Ok)
        self.gridLayout.addWidget(self.buttonBox, 1, 1, 2, 1)

        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.setWindowTitle('Set parameter range')

        self.options = options
        self.comboBox.addItems(options)
        self.comboBox.activated.connect(self.addToList)
        self.lineEdit.setText(' '.join(options))
        self.lineEdit.keyPressEvent = self.keyPressEvent

    def addToList(self, option):
        self.lineEdit.setText(str(self.lineEdit.text()) + ' ' + self.options[option])

    def keyPressEvent(self, ev):
        if ev.key() == QtCore.Qt.Key_Backspace or ev.key() == QtCore.Qt.Key_Delete:
            self.lineEdit.setText(' '.join(str(self.lineEdit.text()).split(' ')[:-1]))
        elif ev.key() == QtCore.Qt.Key_Enter or ev.key() == QtCore.Qt.Key_Return:
            self.addToList(self.comboBox.currentIndex())
        ev.accept()

    def selectedRange(self):
        return str(self.lineEdit.text()).split(' ')






















from collections import OrderedDict
import yaml


class UnsortableList(list):
    def sort(self, *args, **kwargs):
        pass


class UnsortableOrderedDict(OrderedDict):
    def items(self, *args, **kwargs):
        return UnsortableList(OrderedDict.items(self, *args, **kwargs))


yaml.add_representer(UnsortableOrderedDict, yaml.representer.SafeRepresenter.represent_dict)


import distributed
from tornado.ioloop import IOLoop
from threading import Thread
from distributed import Scheduler, Worker, Executor

import paramiko
import select

import getpass
import os
import socket
import select
import threading
from pipeline import msg

try:
    import SocketServer
except ImportError:
    import socketserver as SocketServer

g_verbose = True


def verbose(s):
    if g_verbose:
        msg.logMessage(s,msg.DEBUG)


class ForwardServer(SocketServer.ThreadingTCPServer):
    daemon_threads = True
    allow_reuse_address = True


class Handler(SocketServer.BaseRequestHandler):
    def handle(self):
        try:
            chan = self.ssh_transport.open_channel('direct-tcpip',
                                                   (self.chain_host, self.chain_port),
                                                   self.request.getpeername())
        except Exception as e:
            verbose('Incoming request to %s:%d failed: %s' % (self.chain_host,
                                                              self.chain_port,
                                                              repr(e)))
            return
        if chan is None:
            verbose('Incoming request to %s:%d was rejected by the SSH server.' %
                    (self.chain_host, self.chain_port))
            return

        verbose('Connected!  Tunnel open %r -> %r -> %r' % (self.request.getpeername(),
                                                            chan.getpeername(), (self.chain_host, self.chain_port)))
        while True:
            r, w, x = select.select([self.request, chan], [], [])
            if self.request in r:
                data = self.request.recv(1024)
                if len(data) == 0:
                    break
                chan.send(data)
            if chan in r:
                data = chan.recv(1024)
                if len(data) == 0:
                    break
                self.request.send(data)

        peername = self.request.getpeername()
        chan.close()
        self.request.close()
        verbose('Tunnel closed from %r' % (peername,))


class DaskScheduler(object):
    def __init__(self, client, ipaddr, port, password, runscript):
        self.client = client
        self.ipaddr = ipaddr
        self.port = port
        self.runscript = runscript
        self.password = password
        # self.command = "/usr/common/graphics/visit/camera/runscript.sh {0} {1}".format(ipaddr, port)
        self.command = runscript + " {0} {1}".format(ipaddr, port)
        msg.logMessage(self.command,msg.DEBUG)

    def serve(self):
        msg.logMessage("Serving: " + self.command,msg.INFO)
        channel = self.client.get_transport().open_session()
        msg.logMessage(channel.get_pty(),msg.DEBUG)
        # channel.exec_command('tty')
        channel.exec_command(self.command)
        readytorun=False
        while True:
            if channel.exit_status_ready():
                break
            rl, wl, xl = select.select([channel], [], [], 10.0)
            if len(rl) > 0:
                line = channel.recv(1024)
                msg.logMessage(line)
                if 'queued and waiting for resources' in line:
                    msg.showMessage('Execution queued and waiting for resources...')
                elif 'has been allocated resources' in line:
                    msg.showMessage('Execution has been allocated resources...')
                    readytorun=True
                if readytorun and line:
                    msg.showMessage('Executing job...')
                    readytorun=False

                if line.find("Password") >= 0:
                    msg.logMessage("writing password",msg.DEBUG)
                    channel.sendall(self.password + "\n")
        msg.logMessage("Ending Dask Scheduler",msg.INFO)


class DaskWorker(object):
    def __init__(self, client, ipaddr, port, nodes, partition, time, procs, threads):
        self.client = client
        self.ipaddr = ipaddr
        self.port = port
        self.nodes = nodes
        self.partition = partition
        self.time = time
        self.procs = procs
        self.threads = threads

        self.command = "/usr/common/graphics/visit/camera/runlocalserver.sh {0} {1}".format(ipaddr, port)
        # self.command = "/usr/common/graphics/visit/camera/runserver.sh {0} {1} {2} {3} {4} {5} {6}".format(ipaddr, port, nodes, partition, time, procs, threads)
        msg.logMessage(self.command,msg.DEBUG)

    def serve(self):
        msg.logMessage("Serving Worker: " + self.command,msg.INFO)
        channel = self.client.get_transport().open_session()
        channel.exec_command(self.command)
        # client.exec_command(self.command)
        while True:
            # if channel.exit_status_ready():
            #  break
            rl, wl, xl = select.select([channel], [], [], 10.0)
            if len(rl) > 0:
                msg.logMessage("rl:", channel.recv(1024),msg.DEBUG)
        msg.logMessage("Ending Dask Worker",msg.INFO)


# client = paramiko.SSHClient()
# client.load_system_host_keys()
# client.connect('edison.nersc.gov', username="hkrishna")

class RemoteScheduler():
    """
       Create a remote executor
    """

    def __init__(self, addr, username, loop, password, machine, runscript):
        self.loop = loop
        self.addr = addr
        self.username = username
        self.password = password
        self.command = runscript
        self.client = paramiko.SSHClient()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.client.load_system_host_keys()
        if len(password) > 0:
            self.client.connect(addr, username=username, password=password)
        else:
            self.client.connect(addr, username=username)
        stdin, stdout, stderr = self.client.exec_command(
            "python -c 'import socket; s=socket.socket(); s.bind((\"\", 0)); result = socket.gethostbyname(socket.gethostname()) + \":\" + str(s.getsockname()[1]); s.close(); print(result)'")
        stdin.close()

        for line in stdout.read().splitlines():
            result = line

        self.remote_addr = result.split(":")
        self.local_port = self.get_free_port()
        if len(machine) > 0:
            self.remote_addr[0] = machine

        msg.logMessage((self.local_port, self.remote_addr),msg.DEBUG)
        self.start_scheduler(self.remote_addr[0], self.remote_addr[1], self.client, password, runscript)
        self.executor = None
        self.forward_tunnel(self.local_port, self.remote_addr[0], self.remote_addr[1], self.client.get_transport())

    def execute(self):
        msg.logMessage("Starting Executor",msg.INFO)
        # self.executor = Executor("{0}:{1}".format(self.remote_addr[0], self.remote_addr[1]))
        self.executor = Executor("{0}:{1}".format("localhost", self.local_port))
        msg.logMessage("End Executor",msg.INFO)

    def close(self):
        self.client.exec_command("killall dask-scheduler dask-worker")
        self.client.exec_command("killall " + os.path.basename(self.command))

    def get_free_port(self):
        s = socket.socket()
        s.bind(("", 0))
        localport = s.getsockname()[1]
        s.close()
        return localport

    """
    def start_slurm_worker(self, ipaddr, port, nodes, partition, time, procs, threads, client):
       self.dask_worker = DaskWorker(client, ipaddr, port, nodes, partition, time, procs, threads)
       server_thread = threading.Thread(target=dask_worker.serve)
       server_thread.daemon = True
       server_thread.start()
    """

    def start_scheduler(self, ipaddr, port, client, password, runscript):
        self.dask_sched = DaskScheduler(client, ipaddr, port, password, runscript)
        server_thread = threading.Thread(target=self.dask_sched.serve)
        # server_thread.daemon = True
        server_thread.start()

    def forward_tunnel(self, local_port, remote_host, remote_port, transport):
        class SubHander(Handler):
            chain_host = remote_host
            chain_port = int(remote_port)
            ssh_transport = transport

        fs = ForwardServer(('', int(local_port)), SubHander)
        msg.logMessage(("FORWARDING: ", local_port, remote_host, remote_port),msg.INFO)
        server_thread = threading.Thread(target=fs.serve_forever)
        server_thread.daemon = True
        server_thread.start()



import base64
import json
import os
from urllib import quote
import time

from client.user import User


class GlobusClient(User):
    """
    Client class to authenticate and use globus online API for dataset
    transfering, moving, deleting, etc.
    """

    AUTH_URL = ("https://nexus.api.globusonline.org/goauth/token?grant_type"
                "=client_credentials")
    TRANSFER_URL = "https://transfer.api.globusonline.org/v0.10"
    STANDARD_ENDPOINTS = ('nersc#edison', 'nersc#cori', 'alsuser#bl832data')
    delete_document = os.path.abspath('./json/globusdelete.json')
    transfer_document = os.path.abspath('./json/globustransfer.json')

    def __init__(self):
        super(GlobusClient, self).__init__()
        self.endpoints = {}
        self.authentication = None

    def login(self, username, password):
        """
        Login to NEWT

        Parameters
        ----------
        username : str
        password : str
        """

        basic_auth = base64.b64encode('%s:%s'
                                      % (username, password))
        headers = {'Authorization': 'Basic %s' % basic_auth}
        r = self.get(self.AUTH_URL, headers=headers)

        if r.ok:
            access_token = r.json()['access_token']
            self.authentication = {'Authorization': 'Globus-Goauthtoken %s'
                                   % access_token}
            return super(GlobusClient, self).login(username)
            return self
        else:
            self.authentication = None
            raise GLOBUSError('Bad Authentication: Unable to log in')

    def add_standard_endpoints(self):
        """
        Add standard endpoints to the endpoint list
        """

        for endpoint in self.STANDARD_ENDPOINTS:
                self.add_endpoint(endpoint)

    def add_endpoint(self, endpoint):
        """
        Append globus endpoint as a key value pair
        where the key is the endpoint name and the value is the response
        in json format

        Parameters
        ----------
        endpoint : str
            endpoint name
        """

        self.endpoints[endpoint] = self.get_endpoint(endpoint)

    def get_endpoint(self, endpoint):
        """
        Get globus endpoint information as json

        Parameters
        ----------
        endpoint : str
            endpoint name
        """
        self.check_login()
        r = self.get(self.TRANSFER_URL + '/endpoint/' + quote(endpoint), headers=self.authentication)

        return self.check_response(r)

    def rm_endpoint(self, endpoint):
        """
        Remove an endpoint from the endpoint list

        Parameters
        ---------
        endpoint : str
            endpoint name
        """

        endpoint = self.endpoints.pop(endpoint)
        del endpoint

    def find_user_endpoints(self):
        """
        Find endpoints owned by user whose canonical name starts with the
        username as in user#*

        Returns
        -------
        list of str
            endpoint names
        """
        self.check_login()
        user_endpoints = []
        r = self.get(self.TRANSFER_URL + '/endpoint_list', headers=self.authentication)

        endpoints = self.check_response(r)
        for endpoint in endpoints['DATA']:
            if endpoint['username'] == self.username:
                user_endpoints.append(endpoint['canonical_name'])

        return user_endpoints

    def rm_endpoint(self, endpoint):
        """
        Remove and endpoint key value pair from endpoint dictionary instance
        variable

        Parameters
        ----------
        endpoint : str
            endpoint name
        """
        self.endpoints.pop(endpoint)

    def is_endpoint_active(self, endpoint):
        """
        Check if an endpoint is active

        Parameters
        ----------
        endpoint : str
            endpoint name

        Returns
        -------
        bool
            true if active, false otherwise
        """

        self.check_login()
        self.add_endpoint(endpoint)
        if self.endpoints[endpoint]['activated'] is True:
            return True
        else:
            return False

    def determine_local_endpoint(self):
        """
        Determine which of the user's personal endpoints is the local machine
        if any. The function will essentially 'touch' a test file and then try
        to look for it through globus to find a match.

        Returns
        -------
        str
            name of local endpoint, None if not found
        """
        self.check_login()
        user_endpoints = self.find_user_endpoints()
        try:
            home = os.environ['HOME']
        except KeyError:
            home = os.environ['HOMEPATH']

        test = os.path.join(home, 'test.spew')
        with open(test, 'wb'):
            os.utime(test, None)

        for endpoint in user_endpoints:
                params = {'path': unicode(test)}
                r = self.get(self.TRANSFER_URL + '/endpoint/' + quote(endpoint) + '/ls',
                             headers=self.authentication, params=params)
                try:
                    file_response = r.json()
                    if 'is a file' in file_response['message']:
                        os.remove(test)
                        return endpoint
                except KeyError:
                    pass

        os.remove(test)
        return None

    def get_dir_contents(self, path, endpoint):
        """
        Get contents of a specified directory in a given endpoint

        Parameters
        ----------
        endpoint : str
            endpoint name
        path : str
            directory path
        dict
            dictionary with contents of directory
        """

        self.check_login()
        self.add_endpoint(endpoint)

        if self.endpoints[endpoint]['activated'] is True:
            params = {'path': path}
            r = self.get(self.TRANSFER_URL + '/endpoint/' + quote(endpoint) + '/ls',
                         headers=self.authentication, params=params)
            directory_items = self.check_response(r)
            directory_items = directory_items['DATA']
            return directory_items
        else:
            raise GLOBUSError('%s could not be activated' % endpoint)

    def transfer_file(self, src_endpoint, src_path, dst_enpoint, dst_path=None):
        """
        Make a file transfer submission to globus

        Parameters
        ----------
        src_endpoint : str
            source endpoint name
        src_path : str
            path to file in source endpoint
        dst_enpoint: str
            destination endpoint name
        dst_path: str
            path to directory in destination endpoint including file name

        Returns
        -------
        json
            globus response
        """

        self.check_login()

        if dst_path is None:
            dst_path = os.path.join(os.path.expanduser('~'), os.path.split(src_path)[-1])

        for endpoint in (src_endpoint, dst_enpoint):
            if not self.is_endpoint_active(endpoint):
                raise GLOBUSError('%s endpoint is not active' % endpoint)

        r = self.get(self.TRANSFER_URL + '/submission_id',
                     headers=self.authentication)
        submission_id = self.check_response(r)

        with open(self.transfer_document) as json_file:
            transfer_submission = json.load(json_file)

        transfer_submission["submission_id"] = submission_id["value"]
        transfer_submission["source_endpoint"] = unicode(src_endpoint)
        transfer_submission["destination_endpoint"] = unicode(dst_enpoint)
        transfer_submission["DATA"][0]["source_path"] = unicode(src_path)
        transfer_submission["DATA"][0]["destination_path"] = unicode(dst_path)
        transfer_submission["label"] = unicode('transfered from SPEW client')

        r = self.post(self.TRANSFER_URL + '/transfer', json=transfer_submission, headers=self.authentication)
        transfer_result = self.check_response(r)

        return transfer_result

    def delete_file(self, endpoint, fpath):
        """
        Make a delete file submission to globus

        Parameters
        ----------
        endpoint: str
            endpoint name
        fpath: str
            path to file to delete

        Returns
        -------
        json
            globus response
        """

        self.check_login()

        if not self.is_endpoint_active(endpoint):
            raise GLOBUSError('%s endpoint is not active' % endpoint)

        r = self.get(self.TRANSFER_URL + '/submission_id', headers=self.authentication)
        submission_id = self.check_response(r)

        with open(self.delete_document) as json_file:
            delete_submission = json.load(json_file)

        delete_submission["submission_id"] = submission_id["value"]
        delete_submission["endpoint"] = unicode(endpoint)
        delete_submission["DATA"][0]["path"] = unicode(fpath)
        delete_submission["label"] = unicode('deleted from SPEW client')

        r = self.post(self.TRANSFER_URL + '/delete', json=delete_submission,
                      headers=self.authentication)
        delete_result = self.check_response(r)

        return delete_result


    def get_file_size(self, path, endpoint):
        """
        Get the size of a file in specified endpoint

        Parameters
        ----------
        path : str
            path to file
        endpoint : str
            endpoint name

        Returns
        -------
        int
            file size in bytes
        """

        self.check_login()

        path, name = os.path.split(path)
        r = self.get_dir_contents(path, endpoint)
        f = filter(lambda i: i['name'] == name, r)[0]

        return f['size']


    def get_task_status(self, task_id):
        """
        Make a GET call to monitor status of a submitted task

        Parameters
        ----------
        id : str
            task id returned from a job submission

        Returns
        -------
        json
            globus response
        """
        self.check_login()

        r = self.get(self.TRANSFER_URL + '/task/' + task_id, headers=self.authentication)

        status = self.check_response(r)
        return status


    def transfer_generator(self, src_endpoint, src_path, dst_enpoint, dst_path=None):
        """
        Sumbits a globus transfer task and yields the fraction downloaded

        Parameters
        ----------
        src_endpoint : str
            source endpoint name
        src_path : str
            path to file in source endpoint
        dst_enpoint : str
            destination endpoint name
        dst_path : str
            path to directory in destination endpoint including file name

        Returns
        -------
        json
            globus response
        """

        self.check_login()
        if dst_path is None:
            dst_path = os.path.join(os.path.expanduser('~'), os.path.split(src_path)[-1])

        size = self.get_file_size(src_path, src_endpoint)
        r = self.transfer_file(src_endpoint, src_path, dst_enpoint, dst_path)
        status = self.get_task_status(r['task_id'])
        while status['status'] == 'ACTIVE':
            time.sleep(3)
            status = self.get_task_status(r['task_id'])
            yield float(status['bytes_transferred'])/float(size)


class GLOBUSError(Exception):
    pass



import numpy as np
from PySide.QtGui import *
from PySide.QtCore import *
import pyqtgraph as pg
from pyqtgraph import parametertree as pt
from astropy.modeling import models, Fittable1DModel, Parameter, fitting


class FitParameter(pt.parameterTypes.GroupParameter):
    sigRangeChanged = Signal()
    def __init__(self,plotwidget):
        super(FitParameter, self).__init__(name='Fitting', type='group')
        self._plotwidget=plotwidget

        self.rangemin = 0
        self.rangemax = 0
        self.showrange = False

        self.rangeROI = pg.LinearRegionItem([0, 1], pg.LinearRegionItem.Vertical, brush=pg.mkBrush([128, 0, 128, 32]))
        for line in self.rangeROI.lines: line.setPen(pg.mkPen(color=[128, 0, 128], width=2))
        self.rangeROI.setVisible(False)
        self.plotwidget.addItem(self.rangeROI)

        self.changeModel(None, models.Lorentz1D)

    @property
    def plotwidget(self):
        return self._plotwidget() if callable(self._plotwidget) else self._plotwidget

    @Slot(object,object)
    def changeModel(self,sender,model):
        self.modelParam = ModelParameter(model, self)
        modeltypeparam = pt.Parameter.create(name='Model', type='list', values=usefulmodels,value=model)
        rangeminparam = pt.Parameter.create(name='Range min.', type='float', value=self.rangemin)
        rangemaxparam = pt.Parameter.create(name='Range max.', type='float', value=self.rangemax)
        showrangeparam = pt.Parameter.create(name='Show range', type='bool', value=self.showrange)

        modeltypeparam.sigValueChanged.connect(self.changeModel)
        self.sigRangeChanged.connect(self.updateROI)
        self.rangeROI.sigRegionChangeFinished.connect(self.updateRange)

        self.clearChildren()
        self.addChildren([modeltypeparam,
                          rangeminparam,
                          rangemaxparam,
                          showrangeparam,
                          self.modelParam])

        self.param('Range min.').sigValueChanged.connect(self.sigRangeChanged)
        self.param('Range max.').sigValueChanged.connect(self.sigRangeChanged)
        self.param('Show range').sigValueChanged.connect(self.sigRangeChanged)


    def updateROI(self):
        self.rangemin = self.param('Range min.').value()
        self.rangemax = self.param('Range max.').value()
        self.showrange = self.param('Show range').value()

        self.rangeROI.sigRegionChangeFinished.disconnect(self.updateRange)
        self.rangeROI.setRegion((self.rangemin, self.rangemax))
        self.rangeROI.setVisible(self.showrange)
        self.rangeROI.sigRegionChangeFinished.connect(self.updateRange)
        self.plotwidget.addItem(self.rangeROI)


    def updateRange(self):
        self.rangemin, self.rangemax = self.rangeROI.getRegion()

        self.param('Range min.').setValue(self.rangemin,blockSignal=self.sigRangeChanged)
        self.param('Range max.').setValue(self.rangemax,blockSignal=self.sigRangeChanged)





class ModelParameter(pt.Parameter):

    def __init__(self,model,parentTree):
        self.parentTree=parentTree
        self.model=model
        inputs={'name':'Inputs','type':'group','children':[]}


        outputs={'name':'Outputs','type':'group','children':[],'readonly':True}
        for param in model.param_names:
            inputs['children'].append({'name':param,'type':'float'})
            outputs['children'].append({'name': param, 'type': 'float'})

        fitAction = {'name':'Run Fit','type':'action'}

        children = [inputs, outputs,fitAction]
        super(ModelParameter, self).__init__(name=model.name,
                                             type='group',
                                             children=children)

        self.param('Run Fit').sigActivated.connect(self.RunFit)



    def RunFit(self):
        for item in self.parentTree.plotwidget.listDataItems():
            if hasattr(item,'isFit'):
                self.parentTree.plotwidget.removeItem(item)
                self.parentTree.plotwidget.removeItem(item.display_text)
                continue
            x,y = item.getData()
            limits = self.parentTree.param('Range min.').value(),self.parentTree.param('Range max.').value()
            y=y[np.logical_and(limits[0]<x,x<limits[1])]
            x=x[np.logical_and(limits[0]<x,x<limits[1])]
            inputs = dict((param.opts['name'],param.value()) for param in self.param('Inputs').children())
            g_init = self.model(**inputs)
            fit_g = fitting.LevMarLSQFitter()
            g = fit_g(g_init, x, y)
            txt=''
            for paramname in g.param_names:
                self.param('Outputs').param(paramname).setValue(getattr(g,paramname).value)
                txt+=paramname +': '+str(getattr(g,paramname).value)+'\n'

            txt=txt[:-1] # remove trailing new line

            # Plot fit in similar style
            pen = pg.mkPen(item.opts['pen'])
            pen.setStyle(Qt.DashLine)
            curve = self.parentTree.plotwidget.plot(x,g(x),pen=pen)
            curve.isFit=True
            curve.display_text = pg.TextItem(text='Test', color=(255, 255, 255), anchor=(0, 1),
                                            fill=pg.mkBrush(255, 127, 0, 100))
            curve.display_text.hide()
            self.parentTree.plotwidget.addItem(curve.display_text)
            def hover(event):
                curve.display_text.setPos(event.pos())
                curve.display_text.setVisible(not event.exit)
                curve.display_text.setText(txt)
            curve.curve.hoverEvent=hover



class LogNormal(models.Gaussian1D):

    @staticmethod
    def evaluate(x, amplitude, mean, stddev):
        return models.Gaussian1D.evaluate(np.log(x),amplitude,np.log(mean),stddev/mean)

    fit_deriv = None


usefulmodels = [models.Gaussian1D,models.Voigt1D,models.Lorentz1D,LogNormal]
usefulmodels = dict((model.name,model) for model in usefulmodels)

## Start Qt event loop unless running in interactive mode.
if __name__ == '__main__':
    # ######## SAVE THIS FOR DEBUGGING SEG FAULTS; issues are usually doing something outside the gui thread
    # import sys
    #
    #
    # def trace(frame, event, arg):
    #     print "%s, %s:%d" % (event, frame.f_code.co_filename, frame.f_lineno)
    #     return trace
    #
    #
    # sys.settrace(trace)

    app = QApplication([])

    ## Create window with ImageView widget
    win = QMainWindow()
    win.resize(800, 800)

    p = pg.PlotWidget()
    range = np.arange(0,50,1)
    p.plot(range,models.Gaussian1D.evaluate(range,10,30,5),pen={'color':'r','width':2})
    p.show()

    fitter = FitParameter(p)
    tree = pt.ParameterTree()
    tree.setParameters(fitter,showTop=False)
    win.setCentralWidget(tree)
    win.setWindowTitle('Fitting')
    win.show()

    QApplication.instance().exec_()



from PySide import QtGui
from PySide import QtCore
from PySide.QtCore import Qt
import os


class newfilewatcher(QtCore.QFileSystemWatcher):
    newFilesDetected = QtCore.Signal(str, list)

    def __init__(self):
        super(newfilewatcher, self).__init__()
        self.childrendict = dict()
        self.directoryChanged.connect(self.checkdirectory)

    def addPath(self, path):
        super(newfilewatcher, self).addPath(path)
        self.childrendict[path] = set(os.listdir(path))

    def checkdirectory(self, path):
        # print(path)
        updatedchildren = set(os.listdir(path))
        newchildren = updatedchildren - self.childrendict[path]
        self.childrendict[path] = updatedchildren
        self.newFilesDetected.emit(path, list(newchildren))



from PySide import QtGui
from PySide import QtCore


class difftoolbar(QtGui.QToolBar):
    sigCake = QtCore.Signal()
    sigRemesh = QtCore.Signal()
    def __init__(self):
        super(difftoolbar, self).__init__()

        self.actionCenterFind = QtGui.QAction(self)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap("xicam/gui/icons_27.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionCenterFind.setIcon(icon1)
        self.actionCenterFind.setObjectName("actionCenterFind")
        self.actionCenterFind.setToolTip('Auto-calibrate AgB')
        self.actionPolyMask = QtGui.QAction(self)
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap("xicam/gui/icons_05.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionPolyMask.setIcon(icon2)
        self.actionPolyMask.setToolTip('Polygon mask')
        self.actionPolyMask.setText("")
        self.actionPolyMask.setObjectName("actionPolyMask")
        self.actionThresholdMask = QtGui.QAction(self)
        icont = QtGui.QIcon()
        icont.addPixmap(QtGui.QPixmap("xicam/gui/icons_03.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionThresholdMask.setIcon(icont)
        self.actionThresholdMask.setToolTip('Threshold mask')
        self.actionThresholdMask.setText("")
        self.actionThresholdMask.setObjectName("actionThresholdMask")
        self.actionThresholdMask.setVisible(False)
        self.actionOpen = QtGui.QAction(self)
        self.actionOpen.setObjectName("actionOpen")
        self.actionOpen.setToolTip('Open')
        self.actionSaveExperiment = QtGui.QAction(self)
        self.actionSaveExperiment.setObjectName("actionSaveExperiment")
        self.actionSaveExperiment.setToolTip('Save experiment')
        self.actionLoadExperiment = QtGui.QAction(self)
        self.actionLoadExperiment.setObjectName("actionLoadExperiment")
        self.actionLoadExperiment.setToolTip('Load experiment')
        self.actionClose = QtGui.QAction(self)
        self.actionClose.setObjectName("actionClose")
        self.actionClose.setToolTip('Close Xi-cam')
        self.actionMasking = QtGui.QAction(self)
        self.actionMasking.setToolTip('Masking')
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap("xicam/gui/icons_03.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionMasking.setIcon(icon3)
        self.actionMasking.setObjectName("actionMasking")
        self.actionLog_Intensity = QtGui.QAction(self)
        self.actionLog_Intensity.setCheckable(True)
        self.actionLog_Intensity.setToolTip('Log intensity')
        icon4 = QtGui.QIcon()
        icon4.addPixmap(QtGui.QPixmap("xicam/gui/icons_02.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionLog_Intensity.setIcon(icon4)
        self.actionLog_Intensity.setObjectName("actionLog_Intensity")
        self.actionLog_Intensity.setChecked(True)
        self.actionCake = QtGui.QAction(self)
        self.actionCake.setToolTip('Cake')
        self.actionCake.setCheckable(True)
        icon5 = QtGui.QIcon()
        icon5.addPixmap(QtGui.QPixmap("xicam/gui/icons_04.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionCake.setIcon(icon5)
        self.actionCake.setObjectName("actionCake")
        self.actionRemove_Cosmics = QtGui.QAction(self)
        self.actionRemove_Cosmics.setToolTip('Remove cosmics')
        icon6 = QtGui.QIcon()
        icon6.addPixmap(QtGui.QPixmap("xicam/gui/icons_06.png"), QtGui.QIcon.Normal, QtGui.QIcon.On)
        self.actionRemove_Cosmics.setIcon(icon6)
        self.actionRemove_Cosmics.setText("")
        self.actionRemove_Cosmics.setObjectName("actionRemove_Cosmics")
        self.actionMaskLoad = QtGui.QAction(self)
        self.actionMaskLoad.setToolTip('Load mask')
        icon7 = QtGui.QIcon()
        icon7.addPixmap(QtGui.QPixmap("xicam/gui/icons_08.png"), QtGui.QIcon.Normal, QtGui.QIcon.On)
        self.actionMaskLoad.setIcon(icon7)
        self.actionMaskLoad.setText("")
        self.actionMaskLoad.setObjectName("actionMaskLoad")
        self.actionRadial_Symmetry = QtGui.QAction(self)
        self.actionRadial_Symmetry.setToolTip('Radial symmetry mask-filling')
        self.actionRadial_Symmetry.setCheckable(True)
        icon15 = QtGui.QIcon()
        icon15.addPixmap(QtGui.QPixmap("xicam/gui/icons_18.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionRadial_Symmetry.setIcon(icon15)
        self.actionRadial_Symmetry.setObjectName("actionRadial_Symmetry")
        self.actionMirror_Symmetry = QtGui.QAction(self)
        self.actionMirror_Symmetry.setToolTip('Mirror symmetry mask-filling')
        self.actionMirror_Symmetry.setCheckable(True)
        icon16 = QtGui.QIcon()
        icon16.addPixmap(QtGui.QPixmap("xicam/gui/icons_17.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionMirror_Symmetry.setIcon(icon16)
        self.actionMirror_Symmetry.setObjectName("actionMirror_Symmetry")
        self.actionShow_Mask = QtGui.QAction(self)
        self.actionShow_Mask.setToolTip('Show mask')
        self.actionShow_Mask.setCheckable(True)
        icon17 = QtGui.QIcon()
        icon17.addPixmap(QtGui.QPixmap("xicam/gui/icons_20.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        icon17.addPixmap(QtGui.QPixmap("xicam/gui/icons_19.png"), QtGui.QIcon.Normal, QtGui.QIcon.On)
        self.actionShow_Mask.setIcon(icon17)
        self.actionShow_Mask.setObjectName("actionShow_Mask")
        self.actionPolygon_Cut = QtGui.QAction(self)
        self.actionPolygon_Cut.setToolTip('Polygon region-of-interest')
        self.actionPolygon_Cut.setCheckable(True)
        icon18 = QtGui.QIcon()
        icon18.addPixmap(QtGui.QPixmap("xicam/gui/icons_21.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionPolygon_Cut.setIcon(icon18)
        self.actionPolygon_Cut.setObjectName("actionPolygon_Cut")
        self.actionVertical_Cut = QtGui.QAction(self)
        self.actionVertical_Cut.setToolTip('Vertical region-of-interest')
        icon19 = QtGui.QIcon()
        icon19.addPixmap(QtGui.QPixmap("xicam/gui/icons_22.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionVertical_Cut.setIcon(icon19)
        self.actionVertical_Cut.setObjectName("actionVertical_Cut")
        self.actionHorizontal_Cut = QtGui.QAction(self)
        self.actionHorizontal_Cut.setToolTip('Horizontal region-of-interest')
        icon20 = QtGui.QIcon()
        icon20.addPixmap(QtGui.QPixmap("xicam/gui/icons_23.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionHorizontal_Cut.setIcon(icon20)
        self.actionHorizontal_Cut.setObjectName("actionHorizontal_Cut")
        self.actionLine_Cut = QtGui.QAction(self)
        self.actionLine_Cut.setToolTip('Line region-of-interest')
        icon21 = QtGui.QIcon()
        icon21.addPixmap(QtGui.QPixmap("xicam/gui/icons_24.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionLine_Cut.setIcon(icon21)
        self.actionLine_Cut.setObjectName("actionLine_Cut")
        self.actionRemeshing = QtGui.QAction(self)
        self.actionRemeshing.setToolTip('GIXS Ewald-sphere correction')
        self.actionRemeshing.setCheckable(True)
        icon23 = QtGui.QIcon()
        icon23.addPixmap(QtGui.QPixmap("xicam/gui/icons_25.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionRemeshing.setIcon(icon23)
        self.actionRemeshing.setObjectName("actionRemeshing")
        self.actionRefine_Center = QtGui.QAction(self)
        self.actionRefine_Center.setToolTip('Refine calibration')
        icon24 = QtGui.QIcon()
        icon24.addPixmap(QtGui.QPixmap("xicam/gui/icons_28.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionRefine_Center.setIcon(icon24)
        self.actionRefine_Center.setObjectName("actionRefine_Center")
        self.actionCalibrate_AgB = QtGui.QAction(self)
        self.actionCalibrate_AgB.setToolTip('Auto-calibrate AgB')
        icon25 = QtGui.QIcon()
        icon25.addPixmap(QtGui.QPixmap("xicam/gui/icons_29.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionCalibrate_AgB.setIcon(icon25)
        self.actionCalibrate_AgB.setObjectName("actionCalibrate_AgB")
        self.actionArc = QtGui.QAction(self)
        self.actionArc.setToolTip('Arc region-of-interest')
        icon26 = QtGui.QIcon()
        icon26.addPixmap(QtGui.QPixmap("xicam/gui/icons_32.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionArc.setIcon(icon26)
        self.actionArc.setObjectName("actionArc")

        self.actionProcess = QtGui.QAction(self)
        self.actionProcess.setToolTip('Process')
        icon27 = QtGui.QIcon()
        icon27.addPixmap(QtGui.QPixmap("xicam/gui/icons_34.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        icon27.addPixmap(QtGui.QPixmap("xicam/gui/icons_33.png"), QtGui.QIcon.Normal, QtGui.QIcon.On)
        self.actionProcess.setIcon(icon27)
        self.actionProcess.setObjectName("actionProcess")
        self.actionProcess.setCheckable(True)
        self.actionProcess.setVisible(False)

        self.actionVideo = QtGui.QAction(self)
        self.actionVideo.setToolTip('Export Video')
        icon28 = QtGui.QIcon()
        icon28.addPixmap(QtGui.QPixmap("xicam/gui/icons_31.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionVideo.setIcon(icon28)
        self.actionVideo.setObjectName("actionVideo")
        self.actionVideo.setVisible(False)

        # self.actionSpaceGroup = QtGui.QAction(self)
        # icon29 = QtGui.QIcon()
        # icon29.addPixmap(QtGui.QPixmap("xicam/gui/icons_35.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        # self.actionSpaceGroup.setIcon(icon29)
        # self.actionSpaceGroup.setObjectName("actionSpaceGroup")
        # self.actionSpaceGroup.setCheckable(True)
        # self.actionSpaceGroup.setVisible(False)

        self.actionCapture = QtGui.QAction(self)
        self.actionCapture.setToolTip('Capture region-of-interest')
        icon30 = QtGui.QIcon()
        icon30.addPixmap(QtGui.QPixmap("xicam/gui/icons_36.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionCapture.setIcon(icon30)
        self.actionCapture.setObjectName("actionCapture")
        self.actionCapture.setVisible(False)

        # self.actionROI = QtGui.QAction(self)
        #icon25 = QtGui.QIcon()
        #icon25.addPixmap(QtGui.QPixmap("xicam/gui/icons_29.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        #self.actionROI.setIcon(icon25)
        #self.actionROI.setObjectName("actionROI")

        menu = QtGui.QMenu()
        # menu.addAction(self.actionShow_Mask)
        menu.addAction(self.actionPolyMask)
        menu.addAction(self.actionRemove_Cosmics)
        menu.addAction(self.actionThresholdMask)
        menu.addAction(self.actionMaskLoad)
        toolbuttonMasking = QtGui.QToolButton()
        toolbuttonMasking.setDefaultAction(self.actionMasking)
        toolbuttonMasking.setMenu(menu)
        toolbuttonMasking.setPopupMode(QtGui.QToolButton.InstantPopup)
        toolbuttonMaskingAction = QtGui.QWidgetAction(self)
        toolbuttonMaskingAction.setDefaultWidget(toolbuttonMasking)

        self.setIconSize(QtCore.QSize(32, 32))

        self.addAction(self.actionProcess)
        self.addAction(self.actionVideo)
        self.addAction(self.actionCalibrate_AgB)
        # self.addAction(self.actionCenterFind)        # Hide old buttons
        #self.addAction(self.actionRefine_Center)
        self.addAction(self.actionShow_Mask)
        self.addAction(toolbuttonMaskingAction)
        self.addAction(self.actionCake)
        self.addAction(self.actionRemeshing)
        self.addAction(self.actionArc)
        self.addAction(self.actionLine_Cut)
        self.addAction(self.actionVertical_Cut)
        self.addAction(self.actionHorizontal_Cut)
        self.addAction(self.actionLog_Intensity)
        self.addAction(self.actionRadial_Symmetry)
        self.addAction(self.actionMirror_Symmetry)
        # self.addAction(self.actionSpaceGroup)
        self.addAction(self.actionCapture)


    def connecttriggers(self, calibrate, centerfind, refine, showmask, cake, remesh, linecut, vertcut, horzcut, logint,
                        radialsym, mirrorsym, roi, arc, polymask, process=None, video=None,
                        capture=None,removecosmics=None,thresholdmask=None):
        self.actionCalibrate_AgB.triggered.connect(calibrate)
        self.actionCenterFind.triggered.connect(centerfind)
        self.actionRefine_Center.triggered.connect(refine)
        self.actionShow_Mask.triggered.connect(showmask)
        self.actionCake.triggered.connect(self.caketoggle)  #####################3
        self.actionRemeshing.triggered.connect(self.remeshtoggle)  ##############
        self.actionLine_Cut.triggered.connect(linecut)
        self.actionVertical_Cut.triggered.connect(vertcut)
        self.actionHorizontal_Cut.triggered.connect(horzcut)
        self.actionPolyMask.triggered.connect(polymask)
        self.actionLog_Intensity.triggered.connect(logint)
        self.actionRadial_Symmetry.triggered.connect(radialsym)
        self.actionMirror_Symmetry.triggered.connect(mirrorsym)
        #self.actionROI.triggered.connect(roi)
        self.actionArc.triggered.connect(arc)

        self.sigCake.connect(cake)
        self.sigRemesh.connect(remesh)

        if process is not None:
            self.actionProcess.setVisible(True)
            self.actionProcess.triggered.connect(process)

        if video is not None:
            self.actionVideo.setVisible(True)
            self.actionVideo.triggered.connect(video)

        # if spacegroup is not None:
        #     self.actionSpaceGroup.setVisible(True)
        #     self.actionSpaceGroup.triggered.connect(spacegroup)

        if capture is not None:
            self.actionCapture.setVisible(True)
            self.actionCapture.triggered.connect(capture)

        if removecosmics is not None:
            self.actionRemove_Cosmics.setVisible(True)
            self.actionRemove_Cosmics.triggered.connect(removecosmics)

        if thresholdmask is not None:
            self.actionThresholdMask.setVisible(True)
            self.actionThresholdMask.triggered.connect(thresholdmask)

    def caketoggle(self):
        if self.actionCake.isChecked():
            self.actionRemeshing.setChecked(False)
        self.sigCake.emit()

    def remeshtoggle(self):
        if self.actionRemeshing.isChecked():
            self.actionCake.setChecked(False)
        self.sigRemesh.emit()




